# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi code` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from functools import partial
import click
import tabulate

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options, arguments
from aiida.cmdline.params.options.commands import code as options_code
from aiida.cmdline.utils import echo
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.cmdline.utils.multi_line_input import ensure_scripts
from aiida.common.exceptions import InputValidationError


@verdi.group('code')
def verdi_code():
    """Setup and manage codes."""


def get_default(key, ctx):
    """
    Get the default argument using a user instance property
    :param value: The name of the property to use
    :param ctx: The click context (which will be used to get the user)
    :return: The default value, or None
    """
    try:
        value = getattr(ctx.code_builder, key)
        if value == '':
            value = None
    except KeyError:
        value = None

    return value


def get_computer_name(ctx):
    return getattr(ctx.code_builder, 'computer').name


def get_on_computer(ctx):
    return not getattr(ctx.code_builder, 'is_local')()


# pylint: disable=unused-argument
def set_code_builder(ctx, param, value):
    """Set the code spec for defaults of following options."""
    from aiida.orm.utils.builders.code import CodeBuilder
    ctx.code_builder = CodeBuilder.from_code(value)
    return value


@verdi_code.command('setup')
@options_code.LABEL()
@options_code.DESCRIPTION()
@options_code.INPUT_PLUGIN()
@options_code.ON_COMPUTER()
@options_code.COMPUTER()
@options_code.REMOTE_ABS_PATH()
@options_code.FOLDER()
@options_code.REL_PATH()
@options.PREPEND_TEXT()
@options.APPEND_TEXT()
@options.NON_INTERACTIVE()
@options.CONFIG_FILE()
@with_dbenv()
def setup_code(non_interactive, **kwargs):
    """Setup a new code."""
    from aiida.common.exceptions import ValidationError
    from aiida.orm.utils.builders.code import CodeBuilder

    if not non_interactive:
        try:
            pre, post = ensure_scripts(kwargs.pop('prepend_text', ''), kwargs.pop('append_text', ''), kwargs)
        except InputValidationError as exception:
            raise click.BadParameter('invalid prepend and or append text: {}'.format(exception))

        kwargs['prepend_text'] = pre
        kwargs['append_text'] = post

    if kwargs.pop('on_computer'):
        kwargs['code_type'] = CodeBuilder.CodeType.ON_COMPUTER
    else:
        kwargs['code_type'] = CodeBuilder.CodeType.STORE_AND_UPLOAD

    code_builder = CodeBuilder(**kwargs)
    code = code_builder.new()

    try:
        code.store()
        code.reveal()
    except ValidationError as exception:
        echo.echo_critical('Unable to store the Code: {}'.format(exception))

    echo.echo_success('Code<{}> {} created'.format(code.pk, code.full_label))


@verdi_code.command('duplicate')
@arguments.CODE(callback=set_code_builder)
@options_code.LABEL(contextual_default=partial(get_default, 'label'))
@options_code.DESCRIPTION(contextual_default=partial(get_default, 'description'))
@options_code.INPUT_PLUGIN(contextual_default=partial(get_default, 'input_plugin'))
@options_code.ON_COMPUTER(contextual_default=get_on_computer)
@options_code.COMPUTER(contextual_default=get_computer_name)
@options_code.REMOTE_ABS_PATH(contextual_default=partial(get_default, 'remote_abs_path'))
@options_code.FOLDER(contextual_default=partial(get_default, 'code_folder'))
@options_code.REL_PATH(contextual_default=partial(get_default, 'code_rel_path'))
@options.PREPEND_TEXT()
@options.APPEND_TEXT()
@options.NON_INTERACTIVE()
@click.option('--hide-original', is_flag=True, default=False, help='Hide the code being copied.')
@click.pass_context
@with_dbenv()
def code_duplicate(ctx, code, non_interactive, **kwargs):
    """Duplicate a code allowing to change some parameters."""
    from aiida.common.exceptions import ValidationError
    from aiida.orm.utils.builders.code import CodeBuilder

    if not non_interactive:
        try:
            pre, post = ensure_scripts(kwargs.pop('prepend_text', ''), kwargs.pop('append_text', ''), kwargs)
        except InputValidationError as exception:
            raise click.BadParameter('invalid prepend and or append text: {}'.format(exception))

        kwargs['prepend_text'] = pre
        kwargs['append_text'] = post

    if kwargs.pop('on_computer'):
        kwargs['code_type'] = CodeBuilder.CodeType.ON_COMPUTER
    else:
        kwargs['code_type'] = CodeBuilder.CodeType.STORE_AND_UPLOAD

    if kwargs.pop('hide_original'):
        code.hide()

    code_builder = ctx.code_builder
    for key, value in kwargs.items():
        if value is not None:
            setattr(code_builder, key, value)
    new_code = code_builder.new()

    try:
        new_code.store()
        new_code.reveal()
    except ValidationError as exception:
        echo.echo_critical('Unable to store the Code: {}'.format(exception))

    echo.echo_success('Code<{}> {} created'.format(new_code.pk, new_code.full_label))


@verdi_code.command()
@arguments.CODE()
@options.VERBOSE()
@with_dbenv()
def show(code, verbose):
    """Display detailed information for a code."""
    click.echo(tabulate.tabulate(code.get_full_text_info(verbose)))


@verdi_code.command()
@arguments.CODES()
@with_dbenv()
def delete(codes):
    """Delete a code.

    Note that it is possible to delete a code only if it has not yet been used
    as an input of a calculation, i.e., if it does not have outgoing links.
    """
    from aiida.common.exceptions import InvalidOperation
    from aiida.orm import Node

    for code in codes:
        try:
            pk = code.pk
            full_label = code.full_label
            Node.objects.delete(pk)  # pylint: disable=no-member
        except InvalidOperation as exception:
            echo.echo_error(str(exception))
        else:
            echo.echo_success('Code<{}> {} deleted'.format(pk, full_label))


@verdi_code.command()
@arguments.CODES()
@with_dbenv()
def hide(codes):
    """Hide one or more codes from `verdi code list`."""
    for code in codes:
        code.hide()
        echo.echo_success('Code<{}> {} hidden'.format(code.pk, code.full_label))


@verdi_code.command()
@arguments.CODES()
@with_dbenv()
def reveal(codes):
    """Reveal one or more hidden codes in `verdi code list`."""
    for code in codes:
        code.reveal()
        echo.echo_success('Code<{}> {} revealed'.format(code.pk, code.full_label))


@verdi_code.command()
@arguments.CODE()
@arguments.LABEL()
@with_dbenv()
def relabel(code, label):
    """Relabel a code."""
    old_label = code.full_label

    try:
        code.relabel(label)
    except InputValidationError as exception:
        echo.echo_critical('invalid code name: {}'.format(exception))
    else:
        echo.echo_success('Code<{}> relabeled from {} to {}'.format(code.pk, old_label, code.full_label))


@verdi_code.command('list')
@options.COMPUTER(help='Filter codes by computer.')
@options.INPUT_PLUGIN(help='Filter codes by calculation input plugin.')
@options.ALL(help='Include hidden codes.')
@options.ALL_USERS(help='Include codes from all users.')
@click.option('-o', '--show-owner', 'show_owner', is_flag=True, default=False, help='Show owners of codes.')
@with_dbenv()
def code_list(computer, input_plugin, all_entries, all_users, show_owner):
    """List the available codes."""
    from aiida.orm import Code  # pylint: disable=redefined-outer-name
    from aiida import orm

    qb_user_filters = dict()
    if not all_users:
        user = orm.User.objects.get_default()
        qb_user_filters['email'] = user.email

    qb_computer_filters = dict()
    if computer is not None:
        qb_computer_filters['name'] = computer.name

    qb_code_filters = dict()
    if input_plugin is not None:
        qb_code_filters['attributes.input_plugin'] = input_plugin.name

    # If not all_entries, hide codes with HIDDEN_KEY extra set to True
    if not all_entries:
        qb_code_filters['or'] = [{
            'extras': {
                '!has_key': Code.HIDDEN_KEY
            }
        }, {
            'extras.{}'.format(Code.HIDDEN_KEY): {
                '==': False
            }
        }]

    echo.echo('# List of configured codes:')
    echo.echo("# (use 'verdi code show CODEID' to see the details)")

    showed_results = False

    # pylint: disable=invalid-name
    if computer is not None:
        qb = orm.QueryBuilder()
        qb.append(Code, tag='code', filters=qb_code_filters, project=['id', 'label'])
        # We have a user assigned to the code so we can ask for the
        # presence of a user even if there is no user filter
        qb.append(orm.User, with_node='code', project=['email'], filters=qb_user_filters)
        # We also add the filter on computer. This will automatically
        # return codes that have a computer (and of course satisfy the
        # other filters). The codes that have a computer attached are the
        # remote codes.
        qb.append(orm.Computer, with_node='code', project=['name'], filters=qb_computer_filters)
        qb.order_by({Code: {'id': 'asc'}})
        showed_results = qb.count() > 0
        print_list_res(qb, show_owner)

    # If there is no filter on computers
    else:
        # Print all codes that have a computer assigned to them
        # (these are the remote codes)
        qb = orm.QueryBuilder()
        qb.append(Code, tag='code', filters=qb_code_filters, project=['id', 'label'])
        # We have a user assigned to the code so we can ask for the
        # presence of a user even if there is no user filter
        qb.append(orm.User, with_node='code', project=['email'], filters=qb_user_filters)
        qb.append(orm.Computer, with_node='code', project=['name'])
        qb.order_by({Code: {'id': 'asc'}})
        print_list_res(qb, show_owner)
        showed_results = showed_results or qb.count() > 0

        # Now print all the local codes. To get the local codes we ask
        # the dbcomputer_id variable to be None.
        qb = orm.QueryBuilder()
        comp_non_existence = {'dbcomputer_id': {'==': None}}
        if not qb_code_filters:
            qb_code_filters = comp_non_existence
        else:
            new_qb_code_filters = {'and': [qb_code_filters, comp_non_existence]}
            qb_code_filters = new_qb_code_filters
        qb.append(Code, tag='code', filters=qb_code_filters, project=['id', 'label'])
        # We have a user assigned to the code so we can ask for the
        # presence of a user even if there is no user filter
        qb.append(orm.User, with_node='code', project=['email'], filters=qb_user_filters)
        qb.order_by({Code: {'id': 'asc'}})
        showed_results = showed_results or qb.count() > 0
        print_list_res(qb, show_owner)
    if not showed_results:
        echo.echo('# No codes found matching the specified criteria.')


def print_list_res(qb_query, show_owner):
    """Print a list of available codes."""
    # pylint: disable=invalid-name
    for tuple_ in qb_query.all():
        if len(tuple_) == 3:
            (pk, label, useremail) = tuple_
            computername = None
        elif len(tuple_) == 4:
            (pk, label, useremail, computername) = tuple_
        else:
            echo.echo_warning('Wrong tuple size')
            return

        if show_owner:
            owner_string = ' ({})'.format(useremail)
        else:
            owner_string = ''
        if computername is None:
            computernamestring = ''
        else:
            computernamestring = '@{}'.format(computername)
        echo.echo('* pk {} - {}{}{}'.format(pk, label, computernamestring, owner_string))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import unittest

from aiida.transports.plugins.local import *

# This will be used by test_all_plugins

plugin_transport = LocalTransport()


class TestGeneric(unittest.TestCase):
    """
    Test whoami on localhost.
    """

    def test_whoami(self):
        import getpass

        with LocalTransport() as t:
            self.assertEquals(t.whoami(), getpass.getuser())


class TestBasicConnection(unittest.TestCase):
    """
    Test basic connections.
    """

    def test_closed_connection(self):
        from aiida.transports.transport import TransportInternalError

        with self.assertRaises(TransportInternalError):
            t = LocalTransport()
            t.listdir()

    def test_invalid_param(self):
        with self.assertRaises(ValueError):
            LocalTransport(unrequired_var='something')

    def test_basic(self):
        with LocalTransport():
            pass


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi comment` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators


@verdi.group('comment')
@decorators.deprecated_command("This command has been deprecated. Please use 'verdi node comment' instead.")
def verdi_comment():
    """Inspect, create and manage node comments."""


@verdi_comment.command()
@decorators.deprecated_command("This command has been deprecated. Please use 'verdi node comment add' instead.")
@options.NODES(required=True)
@click.argument('content', type=click.STRING, required=False)
@click.pass_context
@decorators.with_dbenv()
def add(ctx, nodes, content):  # pylint: disable=too-many-arguments, unused-argument
    """Add a comment to one or more nodes."""
    from aiida.cmdline.commands.cmd_node import comment_add
    ctx.forward(comment_add)


@verdi_comment.command()
@decorators.deprecated_command("This command has been deprecated. Please use 'verdi node comment update' instead.")
@click.argument('comment_id', type=int, metavar='COMMENT_ID')
@click.argument('content', type=click.STRING, required=False)
@click.pass_context
@decorators.with_dbenv()
def update(ctx, comment_id, content):  # pylint: disable=too-many-arguments, unused-argument
    """Update a comment of a node."""
    from aiida.cmdline.commands.cmd_node import comment_update
    ctx.forward(comment_update)


@verdi_comment.command()
@decorators.deprecated_command("This command has been deprecated. Please use 'verdi node comment show' instead.")
@options.USER()
@arguments.NODES()
@click.pass_context
@decorators.with_dbenv()
def show(ctx, user, nodes):  # pylint: disable=too-many-arguments, unused-argument
    """Show the comments of one or multiple nodes."""
    from aiida.cmdline.commands.cmd_node import comment_show
    ctx.forward(comment_show)


@verdi_comment.command()
@decorators.deprecated_command("This command has been deprecated. Please use 'verdi node comment remove' instead.")
@options.FORCE()
@click.argument('comment', type=int, required=True, metavar='COMMENT_ID')
@click.pass_context
@decorators.with_dbenv()
def remove(ctx, force, comment):  # pylint: disable=too-many-arguments, unused-argument
    """Remove a comment of a node."""
    from aiida.cmdline.commands.cmd_node import comment_remove
    ctx.forward(comment_remove)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,missing-docstring,too-many-lines
"""Tests for the QueryBuilder."""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import warnings

from six.moves import range, zip

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common.links import LinkType
from aiida.manage import configuration


class TestQueryBuilder(AiidaTestCase):

    def setUp(self):
        super(TestQueryBuilder, self).setUp()
        self.clean_db()
        self.insert_data()

    def test_ormclass_type_classification(self):
        """
        This tests the classifications of the QueryBuilder
        """
        # pylint: disable=protected-access
        from aiida.common.exceptions import DbContentError

        qb = orm.QueryBuilder()

        # Asserting that improper declarations of the class type raise an error
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, 'data')
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, 'data.Data')
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, '.')

        # Asserting that the query type string and plugin type string are returned:
        for _cls, classifiers in (
            qb._get_ormclass(orm.StructureData, None),
            qb._get_ormclass(None, 'data.structure.StructureData.'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], orm.StructureData._plugin_type_string)  # pylint: disable=no-member

        for _cls, classifiers in (
            qb._get_ormclass(orm.Group, None),
            qb._get_ormclass(None, 'group'),
            qb._get_ormclass(None, 'Group'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'group')

        for _cls, classifiers in (
            qb._get_ormclass(orm.User, None),
            qb._get_ormclass(None, 'user'),
            qb._get_ormclass(None, 'User'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'user')

        for _cls, classifiers in (
            qb._get_ormclass(orm.Computer, None),
            qb._get_ormclass(None, 'computer'),
            qb._get_ormclass(None, 'Computer'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'computer')

        for _cls, classifiers in (
            qb._get_ormclass(orm.Data, None),
            qb._get_ormclass(None, 'data.Data.'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], orm.Data._plugin_type_string)  # pylint: disable=no-member

    def test_process_type_classification(self):
        """
        This tests the classifications of the QueryBuilder
        """
        from aiida.engine import WorkChain
        from aiida.plugins import CalculationFactory

        ArithmeticAdd = CalculationFactory('arithmetic.add')

        qb = orm.QueryBuilder()

        # pylint: disable=protected-access

        # When passing a WorkChain class, it should return the type of the corresponding Node
        # including the appropriate filter on the process_type
        _cls, classifiers = qb._get_ormclass(WorkChain, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.workflow.workchain.WorkChainNode.')
        self.assertEqual(classifiers['process_type_string'], 'aiida.engine.processes.workchains.workchain.WorkChain')

        # When passing a WorkChainNode, no process_type filter is applied
        _cls, classifiers = qb._get_ormclass(orm.WorkChainNode, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.workflow.workchain.WorkChainNode.')
        self.assertEqual(classifiers['process_type_string'], None)

        # Same tests for a calculation
        _cls, classifiers = qb._get_ormclass(ArithmeticAdd, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.calculation.calcjob.CalcJobNode.')
        self.assertEqual(classifiers['process_type_string'], 'aiida.calculations:arithmetic.add')

    def test_process_query(self):
        """
        Test querying for a process class.
        """
        from aiida.engine import run, WorkChain, if_, return_, ExitCode
        from aiida.common.warnings import AiidaEntryPointWarning

        class PotentialFailureWorkChain(WorkChain):
            EXIT_STATUS = 1
            EXIT_MESSAGE = 'Well you did ask for it'
            OUTPUT_LABEL = 'optional_output'
            OUTPUT_VALUE = 144

            @classmethod
            def define(cls, spec):
                super(PotentialFailureWorkChain, cls).define(spec)
                spec.input('success', valid_type=orm.Bool)
                spec.input('through_return', valid_type=orm.Bool, default=orm.Bool(False))
                spec.input('through_exit_code', valid_type=orm.Bool, default=orm.Bool(False))
                spec.exit_code(cls.EXIT_STATUS, 'EXIT_STATUS', cls.EXIT_MESSAGE)
                spec.outline(if_(cls.should_return_out_of_outline)(return_(cls.EXIT_STATUS)), cls.failure, cls.success)
                spec.output(cls.OUTPUT_LABEL, required=False)

            def should_return_out_of_outline(self):
                return self.inputs.through_return.value

            def failure(self):
                # pylint: disable=no-else-return

                if self.inputs.success.value is False:
                    # Returning either 0 or ExitCode with non-zero status should terminate the workchain
                    if self.inputs.through_exit_code.value is False:
                        return self.EXIT_STATUS
                    else:
                        return self.exit_codes.EXIT_STATUS  # pylint: disable=no-member
                else:
                    # Returning 0 or ExitCode with zero status should *not* terminate the workchain
                    if self.inputs.through_exit_code.value is False:
                        return 0
                    else:
                        return ExitCode()

            def success(self):
                self.out(self.OUTPUT_LABEL, orm.Int(self.OUTPUT_VALUE).store())

        class DummyWorkChain(WorkChain):
            pass

        # Run a simple test WorkChain
        _result = run(PotentialFailureWorkChain, success=orm.Bool(True))

        # Query for nodes associated with this type of WorkChain
        qb = orm.QueryBuilder()

        with warnings.catch_warnings(record=True) as w:  # pylint: disable=no-member
            # Cause all warnings to always be triggered.
            warnings.simplefilter('always')  # pylint: disable=no-member

            qb.append(PotentialFailureWorkChain)

            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, AiidaEntryPointWarning)

        # There should be one result of type WorkChainNode
        self.assertEqual(qb.count(), 1)
        self.assertTrue(isinstance(qb.all()[0][0], orm.WorkChainNode))

        # Query for nodes of a different type of WorkChain
        qb = orm.QueryBuilder()

        with warnings.catch_warnings(record=True) as w:  # pylint: disable=no-member
            # Cause all warnings to always be triggered.
            warnings.simplefilter('always')  # pylint: disable=no-member

            qb.append(DummyWorkChain)

            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, AiidaEntryPointWarning)

        # There should be no result
        self.assertEqual(qb.count(), 0)

        # Query for all WorkChain nodes
        qb = orm.QueryBuilder()
        qb.append(WorkChain)

        # There should be one result
        self.assertEqual(qb.count(), 1)

    def test_simple_query_1(self):
        """
        Testing a simple query
        """
        # pylint: disable=too-many-statements

        n1 = orm.Data()
        n1.label = 'node1'
        n1.set_attribute('foo', ['hello', 'goodbye'])
        n1.store()

        n2 = orm.CalculationNode()
        n2.label = 'node2'
        n2.set_attribute('foo', 1)

        n3 = orm.Data()
        n3.label = 'node3'
        n3.set_attribute('foo', 1.0000)  # Stored as fval
        n3.store()

        n4 = orm.CalculationNode()
        n4.label = 'node4'
        n4.set_attribute('foo', 'bar')

        n5 = orm.Data()
        n5.label = 'node5'
        n5.set_attribute('foo', None)
        n5.store()

        n2.add_incoming(n1, link_type=LinkType.INPUT_CALC, link_label='link1')
        n2.store()
        n3.add_incoming(n2, link_type=LinkType.CREATE, link_label='link2')

        n4.add_incoming(n3, link_type=LinkType.INPUT_CALC, link_label='link3')
        n4.store()
        n5.add_incoming(n4, link_type=LinkType.CREATE, link_label='link4')

        qb1 = orm.QueryBuilder()
        qb1.append(orm.Node, filters={'attributes.foo': 1.000})

        self.assertEqual(len(qb1.all()), 2)

        qb2 = orm.QueryBuilder()
        qb2.append(orm.Data)
        self.assertEqual(qb2.count(), 3)

        qb2 = orm.QueryBuilder()
        qb2.append(entity_type='data.Data.')
        self.assertEqual(qb2.count(), 3)

        qb3 = orm.QueryBuilder()
        qb3.append(orm.Node, project='label', tag='node1')
        qb3.append(orm.Node, project='label', tag='node2')
        self.assertEqual(qb3.count(), 4)

        qb4 = orm.QueryBuilder()
        qb4.append(orm.CalculationNode, tag='node1')
        qb4.append(orm.Data, tag='node2')
        self.assertEqual(qb4.count(), 2)

        qb5 = orm.QueryBuilder()
        qb5.append(orm.Data, tag='node1')
        qb5.append(orm.CalculationNode, tag='node2')
        self.assertEqual(qb5.count(), 2)

        qb6 = orm.QueryBuilder()
        qb6.append(orm.Data, tag='node1')
        qb6.append(orm.Data, tag='node2')
        self.assertEqual(qb6.count(), 0)

    def test_simple_query_2(self):
        from datetime import datetime
        from aiida.common.exceptions import MultipleObjectsError, NotExistent
        n0 = orm.Data()
        n0.label = 'hello'
        n0.description = ''
        n0.set_attribute('foo', 'bar')

        n1 = orm.CalculationNode()
        n1.label = 'foo'
        n1.description = 'I am FoO'

        n2 = orm.Data()
        n2.label = 'bar'
        n2.description = 'I am BaR'

        n2.add_incoming(n1, link_type=LinkType.CREATE, link_label='random_2')
        n1.add_incoming(n0, link_type=LinkType.INPUT_CALC, link_label='random_1')

        for n in (n0, n1, n2):
            n.store()

        qb1 = orm.QueryBuilder()
        qb1.append(orm.Node, filters={'label': 'hello'})
        self.assertEqual(len(list(qb1.all())), 1)

        qh = {
            'path': [{
                'cls': orm.Node,
                'tag': 'n1'
            }, {
                'cls': orm.Node,
                'tag': 'n2',
                'with_incoming': 'n1'
            }],
            'filters': {
                'n1': {
                    'label': {
                        'ilike': '%foO%'
                    },
                },
                'n2': {
                    'label': {
                        'ilike': 'bar%'
                    },
                }
            },
            'project': {
                'n1': ['id', 'uuid', 'ctime', 'label'],
                'n2': ['id', 'description', 'label'],
            }
        }

        qb2 = orm.QueryBuilder(**qh)

        resdict = qb2.dict()
        self.assertEqual(len(resdict), 1)
        self.assertTrue(isinstance(resdict[0]['n1']['ctime'], datetime))

        res_one = qb2.one()
        self.assertTrue('bar' in res_one)

        qh = {
            'path': [{
                'cls': orm.Node,
                'tag': 'n1'
            }, {
                'cls': orm.Node,
                'tag': 'n2',
                'with_incoming': 'n1'
            }],
            'filters': {
                'n1--n2': {
                    'label': {
                        'like': '%_2'
                    }
                }
            }
        }
        qb = orm.QueryBuilder(**qh)
        self.assertEqual(qb.count(), 1)

        # Test the hashing:
        query1 = qb.get_query()
        qb.add_filter('n2', {'label': 'nonexistentlabel'})
        self.assertEqual(qb.count(), 0)

        with self.assertRaises(NotExistent):
            qb.one()
        with self.assertRaises(MultipleObjectsError):
            orm.QueryBuilder().append(orm.Node).one()

        query2 = qb.get_query()
        query3 = qb.get_query()

        self.assertTrue(id(query1) != id(query2))
        self.assertTrue(id(query2) == id(query3))

    def test_operators_eq_lt_gt(self):
        nodes = [orm.Data() for _ in range(8)]

        nodes[0].set_attribute('fa', 1)
        nodes[1].set_attribute('fa', 1.0)
        nodes[2].set_attribute('fa', 1.01)
        nodes[3].set_attribute('fa', 1.02)
        nodes[4].set_attribute('fa', 1.03)
        nodes[5].set_attribute('fa', 1.04)
        nodes[6].set_attribute('fa', 1.05)
        nodes[7].set_attribute('fa', 1.06)

        for n in nodes:
            n.store()

        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<': 1}}).count(), 0)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'==': 1}}).count(), 2)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<': 1.02}}).count(), 3)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<=': 1.02}}).count(), 4)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'>': 1.02}}).count(), 4)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'>=': 1.02}}).count(), 5)

    def test_subclassing(self):
        s = orm.StructureData()
        s.set_attribute('cat', 'miau')
        s.store()

        d = orm.Data()
        d.set_attribute('cat', 'miau')
        d.store()

        p = orm.Dict(dict=dict(cat='miau'))
        p.store()

        # Now when asking for a node with attr.cat==miau, I want 3 esults:
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 3)

        qb = orm.QueryBuilder().append(orm.Data, filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 3)

        # If I'm asking for the specific lowest subclass, I want one result
        for cls in (orm.StructureData, orm.Dict):
            qb = orm.QueryBuilder().append(cls, filters={'attributes.cat': 'miau'})
            self.assertEqual(qb.count(), 1)

        # Now I am not allow the subclassing, which should give 1 result for each
        for cls, count in ((orm.StructureData, 1), (orm.Dict, 1), (orm.Data, 1), (orm.Node, 0)):
            qb = orm.QueryBuilder().append(cls, filters={'attributes.cat': 'miau'}, subclassing=False)
            self.assertEqual(qb.count(), count)

        # Now I am testing the subclassing with tuples:
        qb = orm.QueryBuilder().append(cls=(orm.StructureData, orm.Dict), filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.dict.Dict.'), filters={'attributes.cat': 'miau'}
        )
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            cls=(orm.StructureData, orm.Dict), filters={'attributes.cat': 'miau'}, subclassing=False
        )
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            cls=(orm.StructureData, orm.Data),
            filters={'attributes.cat': 'miau'},
        )
        self.assertEqual(qb.count(), 3)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.dict.Dict.'),
            filters={'attributes.cat': 'miau'},
            subclassing=False
        )
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.Data.'),
            filters={'attributes.cat': 'miau'},
            subclassing=False
        )
        self.assertEqual(qb.count(), 2)

    def test_list_behavior(self):
        for _i in range(4):
            orm.Data().store()

        self.assertEqual(len(orm.QueryBuilder().append(orm.Node).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project='*').all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['id']).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node).dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project='*').dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['id']).dict()), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project='*').iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['id']).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node).iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project='*').iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['id']).iterdict())), 4)

    def test_append_validation(self):
        from aiida.common.exceptions import InputValidationError

        # So here I am giving two times the same tag
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, tag='n').append(orm.StructureData, tag='n')
        # here I am giving a wrong filter specifications
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, filters=['jajjsd'])
        # here I am giving a nonsensical projection:
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, project=True)

        # here I am giving a nonsensical projection for the edge:
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.ProcessNode).append(orm.StructureData, edge_tag='t').add_projection('t', True)
        # Giving a nonsensical limit
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.ProcessNode).limit(2.3)
        # Giving a nonsensical offset
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder(offset=2.3)

        # So, I mess up one append, I want the QueryBuilder to clean it!
        with self.assertRaises(InputValidationError):
            qb = orm.QueryBuilder()
            # This also checks if we correctly raise for wrong keywords
            qb.append(orm.StructureData, tag='s', randomkeyword={})

        # Now I'm checking whether this keyword appears anywhere in the internal dictionaries:
        # pylint: disable=protected-access
        self.assertTrue('s' not in qb._projections)
        self.assertTrue('s' not in qb._filters)
        self.assertTrue('s' not in qb.tag_to_alias_map)
        self.assertTrue(len(qb._path) == 0)
        self.assertTrue(orm.StructureData not in qb._cls_to_tag_map)
        # So this should work now:
        qb.append(orm.StructureData, tag='s').limit(2).dict()

    def test_tags(self):
        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='n1')
        qb.append(orm.Node, tag='n2', edge_tag='e1', with_incoming='n1')
        qb.append(orm.Node, tag='n3', edge_tag='e2', with_incoming='n2')
        qb.append(orm.Computer, with_node='n3', tag='c1', edge_tag='nonsense')
        self.assertEqual(qb.get_used_tags(), ['n1', 'n2', 'e1', 'n3', 'e2', 'c1', 'nonsense'])

        # Now I am testing the default tags,
        qb = orm.QueryBuilder().append(orm.StructureData
                                      ).append(orm.ProcessNode).append(orm.StructureData
                                                                      ).append(orm.Dict, with_outgoing=orm.ProcessNode)
        self.assertEqual(
            qb.get_used_tags(), [
                'StructureData_1', 'ProcessNode_1', 'StructureData_1--ProcessNode_1', 'StructureData_2',
                'ProcessNode_1--StructureData_2', 'Dict_1', 'ProcessNode_1--Dict_1'
            ]
        )
        self.assertEqual(
            qb.get_used_tags(edges=False), [
                'StructureData_1',
                'ProcessNode_1',
                'StructureData_2',
                'Dict_1',
            ]
        )
        self.assertEqual(
            qb.get_used_tags(vertices=False),
            ['StructureData_1--ProcessNode_1', 'ProcessNode_1--StructureData_2', 'ProcessNode_1--Dict_1']
        )
        self.assertEqual(
            qb.get_used_tags(edges=False), [
                'StructureData_1',
                'ProcessNode_1',
                'StructureData_2',
                'Dict_1',
            ]
        )
        self.assertEqual(
            qb.get_used_tags(vertices=False),
            ['StructureData_1--ProcessNode_1', 'ProcessNode_1--StructureData_2', 'ProcessNode_1--Dict_1']
        )


class TestQueryHelp(AiidaTestCase):

    def test_queryhelp(self):
        """
        Here I test the queryhelp by seeing whether results are the same as using the append method.
        I also check passing of tuples.
        """
        g = orm.Group(label='helloworld').store()
        for cls in (orm.StructureData, orm.Dict, orm.Data):
            obj = cls()
            obj.set_attribute('foo-qh2', 'bar')
            obj.store()
            g.add_nodes(obj)

        for cls, expected_count, subclassing in (
            (orm.StructureData, 1, True),
            (orm.Dict, 1, True),
            (orm.Data, 3, True),
            (orm.Data, 1, False),
            ((orm.Dict, orm.StructureData), 2, True),
            ((orm.Dict, orm.StructureData), 2, False),
            ((orm.Dict, orm.Data), 2, False),
            ((orm.Dict, orm.Data), 3, True),
            ((orm.Dict, orm.Data, orm.StructureData), 3, False),
        ):
            qb = orm.QueryBuilder()
            qb.append(cls, filters={'attributes.foo-qh2': 'bar'}, subclassing=subclassing, project='uuid')
            self.assertEqual(qb.count(), expected_count)

            qh = qb.get_json_compatible_queryhelp()
            qb_new = orm.QueryBuilder(**qh)
            self.assertEqual(qb_new.count(), expected_count)
            self.assertEqual(sorted([uuid for uuid, in qb.all()]), sorted([uuid for uuid, in qb_new.all()]))

        qb = orm.QueryBuilder().append(orm.Group, filters={'label': 'helloworld'})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append((orm.Group,), filters={'label': 'helloworld'})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append(orm.Computer,)
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append(cls=(orm.Computer,))
        self.assertEqual(qb.count(), 1)


class TestQueryBuilderCornerCases(AiidaTestCase):
    """
    In this class corner cases of QueryBuilder are added.
    """

    def test_computer_json(self):  # pylint: disable=no-self-use
        """
        In this test we check the correct behavior of QueryBuilder when
        retrieving the _metadata with no content.
        Note that they are in JSON format in both backends. Forcing the
        decoding of a None value leads to an exception (this was the case
        under Django).
        """
        n1 = orm.CalculationNode()
        n1.label = 'node2'
        n1.set_attribute('foo', 1)
        n1.store()

        # Checking the correct retrieval of _metadata which is
        # a JSON field (in both backends).
        qb = orm.QueryBuilder()
        qb.append(orm.CalculationNode, project=['id'], tag='calc')
        qb.append(orm.Computer, project=['id', 'metadata'], outerjoin=True, with_node='calc')
        qb.all()


class TestAttributes(AiidaTestCase):

    def test_attribute_existence(self):
        # I'm storing a value under key whatever:
        val = 1.
        res_uuids = set()
        n1 = orm.Data()
        n1.set_attribute('whatever', 3.)
        n1.set_attribute('test_case', 'test_attribute_existence')
        n1.store()

        # I want all the nodes where whatever is smaller than 1. or there is no such value:

        qb = orm.QueryBuilder()
        qb.append(
            orm.Data,
            filters={
                'or': [{
                    'attributes': {
                        '!has_key': 'whatever'
                    }
                }, {
                    'attributes.whatever': {
                        '<': val
                    }
                }],
            },
            project='uuid'
        )
        res_query = {str(_[0]) for _ in qb.all()}
        self.assertEqual(res_query, res_uuids)

    def test_attribute_type(self):
        key = 'value_test_attr_type'
        n_int, n_float, n_str, n_str2, n_bool, n_arr = [orm.Data() for _ in range(6)]
        n_int.set_attribute(key, 1)
        n_float.set_attribute(key, 1.0)
        n_bool.set_attribute(key, True)
        n_str.set_attribute(key, '1')
        n_str2.set_attribute(key, 'one')
        n_arr.set_attribute(key, [4, 3, 5])

        for n in (n_str2, n_str, n_int, n_float, n_bool, n_arr):
            n.store()

        # Here I am testing which values contain a number 1.
        # Both 1 and 1.0 are legitimate values if ask for either 1 or 1.0
        for val in (1.0, 1):
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): val}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'>': 0.5}}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'<': 1.5}}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
        # Now I am testing the boolean value:
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): True}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_bool.uuid,)))

        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'like': '%n%'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str2.uuid,)))
        qb = orm.QueryBuilder().append(
            orm.Node, filters={'attributes.{}'.format(key): {
                                   'ilike': 'On%'
                               }}, project='uuid'
        )
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str2.uuid,)))
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'like': '1'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str.uuid,)))
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'==': '1'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str.uuid,)))
        if configuration.PROFILE.database_backend == u'sqlalchemy':
            # I can't query the length of an array with Django,
            # so I exclude. Not the nicest way, But I would like to keep this piece
            # of code because of the initialization part, that would need to be
            # duplicated or wrapped otherwise.
            qb = orm.QueryBuilder().append(
                orm.Node, filters={'attributes.{}'.format(key): {
                                       'of_length': 3
                                   }}, project='uuid'
            )
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_arr.uuid,)))


class QueryBuilderLimitOffsetsTest(AiidaTestCase):

    def test_ordering_limits_offsets_of_results_general(self):
        # Creating 10 nodes with an attribute that can be ordered
        for i in range(10):
            n = orm.Data()
            n.set_attribute('foo', i)
            n.store()

        qb = orm.QueryBuilder().append(orm.Node, project='attributes.foo').order_by({orm.Node: 'ctime'})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(10)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 10)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 8)))

        # Specifying the order  explicitly the order:
        qb = orm.QueryBuilder().append(orm.Node,
                                       project='attributes.foo').order_by({orm.Node: {
                                           'ctime': {
                                               'order': 'asc'
                                           }
                                       }})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(10)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 10)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 8)))

        # Reversing the order:
        qb = orm.QueryBuilder().append(orm.Node,
                                       project='attributes.foo').order_by({orm.Node: {
                                           'ctime': {
                                               'order': 'desc'
                                           }
                                       }})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(9, -1, -1)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(4, -1, -1)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(4, 1, -1)))


class QueryBuilderJoinsTests(AiidaTestCase):

    def test_joins1(self):
        # Creating n1, who will be a parent:
        parent = orm.Data()
        parent.label = 'mother'
        parent.store()

        good_child = orm.CalculationNode()
        good_child.label = 'good_child'
        good_child.set_attribute('is_good', True)

        bad_child = orm.CalculationNode()
        bad_child.label = 'bad_child'
        bad_child.set_attribute('is_good', False)

        unrelated = orm.CalculationNode()
        unrelated.label = 'unrelated'
        unrelated.store()

        good_child.add_incoming(parent, link_type=LinkType.INPUT_CALC, link_label='parent')
        bad_child.add_incoming(parent, link_type=LinkType.INPUT_CALC, link_label='parent')
        good_child.store()
        bad_child.store()

        # Using a standard inner join
        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='parent')
        qb.append(orm.Node, tag='children', project='label', filters={'attributes.is_good': True})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='parent')
        qb.append(orm.Node, tag='children', outerjoin=True, project='label', filters={'attributes.is_good': True})
        self.assertEqual(qb.count(), 1)

    def test_joins2(self):
        # Creating n1, who will be a parent:

        students = [orm.Data() for i in range(10)]
        advisors = [orm.CalculationNode() for i in range(3)]
        for i, a in enumerate(advisors):
            a.label = 'advisor {}'.format(i)
            a.set_attribute('advisor_id', i)

        for n in advisors + students:
            n.store()

        # advisor 0 get student 0, 1
        for i in (0, 1):
            students[i].add_incoming(advisors[0], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # advisor 1 get student 3, 4
        for i in (3, 4):
            students[i].add_incoming(advisors[1], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # advisor 2 get student 5, 6, 7
        for i in (5, 6, 7):
            students[i].add_incoming(advisors[2], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # let's add a differnt relationship than advisor:
        students[9].add_incoming(advisors[2], link_type=LinkType.CREATE, link_label='lover')

        self.assertEqual(
            orm.QueryBuilder().append(
                orm.Node
            ).append(orm.Node, edge_filters={
                'label': {
                    'like': 'is\\_advisor\\_%'
                }
            }, tag='student').count(), 7
        )

        for adv_id, number_students in zip(list(range(3)), (2, 2, 3)):
            self.assertEqual(
                orm.QueryBuilder().append(orm.Node, filters={
                    'attributes.advisor_id': adv_id
                }).append(orm.Node, edge_filters={
                    'label': {
                        'like': 'is\\_advisor\\_%'
                    }
                }, tag='student').count(), number_students
            )

    def test_joins3_user_group(self):
        # Create another user
        new_email = 'newuser@new.n'
        user = orm.User(email=new_email).store()

        # Create a group that belongs to that user
        group = orm.Group(label='node_group')
        group.user = user
        group.store()

        # Search for the group of the user
        qb = orm.QueryBuilder()
        qb.append(orm.User, tag='user', filters={'id': {'==': user.id}})
        qb.append(orm.Group, with_user='user', filters={'id': {'==': group.id}})
        self.assertEqual(qb.count(), 1, 'The expected group that belongs to ' 'the selected user was not found.')

        # Search for the user that owns a group
        qb = orm.QueryBuilder()
        qb.append(orm.Group, tag='group', filters={'id': {'==': group.id}})
        qb.append(orm.User, with_group='group', filters={'id': {'==': user.id}})

        self.assertEqual(qb.count(), 1, 'The expected user that owns the ' 'selected group was not found.')

    def test_joins_group_node(self):
        """
        This test checks that the querying for the nodes that belong to a group works correctly (using QueryBuilder).
        This is important for the Django backend with the use of aldjemy for the Django to SQLA schema translation.
        Since this is not backend specific test (even if it is mainly used to test the querying of Django backend
        with QueryBuilder), we keep it at the general tests (ran by both backends).
        """
        new_email = 'newuser@new.n2'
        user = orm.User(email=new_email).store()

        # Create a group that belongs to that user
        group = orm.Group(label='node_group_2')
        group.user = user
        group.store()

        # Create nodes and add them to the created group
        n1 = orm.Data()
        n1.label = 'node1'
        n1.set_attribute('foo', ['hello', 'goodbye'])
        n1.store()

        n2 = orm.CalculationNode()
        n2.label = 'node2'
        n2.set_attribute('foo', 1)
        n2.store()

        n3 = orm.Data()
        n3.label = 'node3'
        n3.set_attribute('foo', 1.0000)  # Stored as fval
        n3.store()

        n4 = orm.CalculationNode()
        n4.label = 'node4'
        n4.set_attribute('foo', 'bar')
        n4.store()

        group.add_nodes([n1, n2, n3, n4])

        # Check that the nodes are in the group
        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='node', project=['id'])
        qb.append(orm.Group, with_node='node', filters={'id': {'==': group.id}})
        self.assertEqual(qb.count(), 4, 'There should be 4 nodes in the group')
        id_res = [_ for [_] in qb.all()]
        for curr_id in [n1.id, n2.id, n3.id, n4.id]:
            self.assertIn(curr_id, id_res)


class QueryBuilderPath(AiidaTestCase):

    def test_query_path(self):
        # pylint: disable=too-many-statements

        q = self.backend.query_manager
        n1 = orm.Data()
        n1.label = 'n1'
        n2 = orm.CalculationNode()
        n2.label = 'n2'
        n3 = orm.Data()
        n3.label = 'n3'
        n4 = orm.Data()
        n4.label = 'n4'
        n5 = orm.CalculationNode()
        n5.label = 'n5'
        n6 = orm.Data()
        n6.label = 'n6'
        n7 = orm.CalculationNode()
        n7.label = 'n7'
        n8 = orm.Data()
        n8.label = 'n8'
        n9 = orm.Data()
        n9.label = 'n9'

        # I create a strange graph, inserting links in a order
        # such that I often have to create the transitive closure
        # between two graphs
        n3.add_incoming(n2, link_type=LinkType.CREATE, link_label='link1')
        n2.add_incoming(n1, link_type=LinkType.INPUT_CALC, link_label='link2')
        n5.add_incoming(n3, link_type=LinkType.INPUT_CALC, link_label='link3')
        n5.add_incoming(n4, link_type=LinkType.INPUT_CALC, link_label='link4')
        n4.add_incoming(n2, link_type=LinkType.CREATE, link_label='link5')
        n7.add_incoming(n6, link_type=LinkType.INPUT_CALC, link_label='link6')
        n8.add_incoming(n7, link_type=LinkType.CREATE, link_label='link7')

        for node in [n1, n2, n3, n4, n5, n6, n7, n8, n9]:
            node.store()

        # There are no parents to n9, checking that
        self.assertEqual(set([]), set(q.get_all_parents([n9.pk])))
        # There is one parent to n6
        self.assertEqual({(_,) for _ in (n6.pk,)}, {tuple(_) for _ in q.get_all_parents([n7.pk])})
        # There are several parents to n4
        self.assertEqual({(_.pk,) for _ in (n1, n2)}, {tuple(_) for _ in q.get_all_parents([n4.pk])})
        # There are several parents to n5
        self.assertEqual({(_.pk,) for _ in (n1, n2, n3, n4)}, {tuple(_) for _ in q.get_all_parents([n5.pk])})

        # Yet, no links from 1 to 8
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 0
        )

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 0
        )

        n6.add_incoming(n5, link_type=LinkType.CREATE, link_label='link1')
        # Yet, now 2 links from 1 to 8
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 2
        )

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 2
        )

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': {
                        '<': 6
                    }
                },
            ).count(), 2
        )
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': 5
                },
            ).count(), 2
        )
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': {
                        '<': 5
                    }
                },
            ).count(), 0
        )

        # TODO write a query that can filter certain paths by traversed ID # pylint: disable=fixme
        qb = orm.QueryBuilder().append(
            orm.Node,
            filters={
                'id': n8.pk
            },
            tag='desc',
        ).append(orm.Node, with_descendants='desc', edge_project='path', filters={'id': n1.pk})
        queried_path_set = {frozenset(p) for p, in qb.all()}

        paths_there_should_be = {
            frozenset([n1.pk, n2.pk, n3.pk, n5.pk, n6.pk, n7.pk, n8.pk]),
            frozenset([n1.pk, n2.pk, n4.pk, n5.pk, n6.pk, n7.pk, n8.pk])
        }

        self.assertTrue(queried_path_set == paths_there_should_be)

        qb = orm.QueryBuilder().append(orm.Node, filters={
            'id': n1.pk
        }, tag='anc').append(orm.Node, with_ancestors='anc', filters={'id': n8.pk}, edge_project='path')

        self.assertEqual({frozenset(p) for p, in qb.all()}, {
            frozenset([n1.pk, n2.pk, n3.pk, n5.pk, n6.pk, n7.pk, n8.pk]),
            frozenset([n1.pk, n2.pk, n4.pk, n5.pk, n6.pk, n7.pk, n8.pk])
        })

        # This part of the test is no longer possible as the nodes have already been stored and the previous parts of
        # the test rely on this, which means however, that here, no more links can be added as that will raise.

        # n7.add_incoming(n9, link_type=LinkType.INPUT_CALC, link_label='link0')
        # # Still two links...

        # self.assertEqual(
        #     orm.QueryBuilder().append(orm.Node, filters={
        #         'id': n1.pk
        #     }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
        #         'id': n8.pk
        #     }).count(), 2)

        # self.assertEqual(
        #     orm.QueryBuilder().append(orm.Node, filters={
        #         'id': n8.pk
        #     }, tag='desc').append(orm.Node, with_descendants='desc', filters={
        #         'id': n1.pk
        #     }).count(), 2)
        # n9.add_incoming(n5, link_type=LinkType.CREATE, link_label='link6')
        # # And now there should be 4 nodes

        # self.assertEqual(
        #     orm.QueryBuilder().append(orm.Node, filters={
        #         'id': n1.pk
        #     }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
        #         'id': n8.pk
        #     }).count(), 4)

        # self.assertEqual(
        #     orm.QueryBuilder().append(orm.Node, filters={
        #         'id': n8.pk
        #     }, tag='desc').append(orm.Node, with_descendants='desc', filters={
        #         'id': n1.pk
        #     }).count(), 4)

        # qb = orm.QueryBuilder().append(
        #     orm.Node, filters={
        #         'id': n1.pk
        #     }, tag='anc').append(
        #         orm.Node, with_ancestors='anc', filters={'id': n8.pk}, edge_tag='edge')
        # qb.add_projection('edge', 'depth')
        # self.assertTrue(set(next(zip(*qb.all()))), set([5, 6]))
        # qb.add_filter('edge', {'depth': 5})
        # self.assertTrue(set(next(zip(*qb.all()))), set([5]))


class TestConsistency(AiidaTestCase):

    def test_create_node_and_query(self):
        """
        Testing whether creating nodes within a iterall iteration changes the results.
        """
        for _i in range(100):
            n = orm.Data()
            n.store()

        for idx, _item in enumerate(
            orm.QueryBuilder().append(orm.Node, project=['id', 'label']).iterall(batch_size=10)
        ):
            if idx % 10 == 10:
                n = orm.Data()
                n.store()
        self.assertEqual(idx, 99)  # pylint: disable=undefined-loop-variable
        self.assertTrue(len(orm.QueryBuilder().append(orm.Node, project=['id', 'label']).all(batch_size=10)) > 99)

    def test_len_results(self):
        """
        Test whether the len of results matches the count returned.
        See also https://github.com/aiidateam/aiida-core/issues/1600
        SQLAlchemy has a deduplication strategy that leads to strange behavior, tested against here
        """
        parent = orm.CalculationNode().store()
        # adding 5 links going out:
        for inode in range(5):
            output_node = orm.Data().store()
            output_node.add_incoming(parent, link_type=LinkType.CREATE, link_label='link_{}'.format(inode))
        for projection in ('id', '*'):
            qb = orm.QueryBuilder()
            qb.append(orm.CalculationNode, filters={'id': parent.id}, tag='parent', project=projection)
            qb.append(orm.Data, with_incoming='parent')
            self.assertEqual(len(qb.all()), qb.count())


class TestManager(AiidaTestCase):

    def test_statistics(self):
        """
        Test if the statistics query works properly.

        I try to implement it in a way that does not depend on the past state.
        """
        from collections import defaultdict

        # pylint: disable=protected-access

        def store_and_add(n, statistics):
            n.store()
            statistics['total'] += 1
            statistics['types'][n._plugin_type_string] += 1  # pylint: disable=no-member
            statistics['ctime_by_day'][n.ctime.strftime('%Y-%m-%d')] += 1

        qmanager = self.backend.query_manager
        current_db_statistics = qmanager.get_creation_statistics()
        types = defaultdict(int)
        types.update(current_db_statistics['types'])
        ctime_by_day = defaultdict(int)
        ctime_by_day.update(current_db_statistics['ctime_by_day'])

        expected_db_statistics = {'total': current_db_statistics['total'], 'types': types, 'ctime_by_day': ctime_by_day}

        store_and_add(orm.Data(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.CalculationNode(), expected_db_statistics)

        new_db_statistics = qmanager.get_creation_statistics()
        # I only check a few fields
        new_db_statistics = {k: v for k, v in new_db_statistics.items() if k in expected_db_statistics}

        expected_db_statistics = {
            k: dict(v) if isinstance(v, defaultdict) else v for k, v in expected_db_statistics.items()
        }

        self.assertEqual(new_db_statistics, expected_db_statistics)

    def test_statistics_default_class(self):
        """
        Test if the statistics query works properly.

        I try to implement it in a way that does not depend on the past state.
        """
        from collections import defaultdict

        def store_and_add(n, statistics):
            n.store()
            statistics['total'] += 1
            statistics['types'][n._plugin_type_string] += 1  # pylint: disable=no-member,protected-access
            statistics['ctime_by_day'][n.ctime.strftime('%Y-%m-%d')] += 1

        current_db_statistics = self.backend.query_manager.get_creation_statistics()
        types = defaultdict(int)
        types.update(current_db_statistics['types'])
        ctime_by_day = defaultdict(int)
        ctime_by_day.update(current_db_statistics['ctime_by_day'])

        expected_db_statistics = {'total': current_db_statistics['total'], 'types': types, 'ctime_by_day': ctime_by_day}

        store_and_add(orm.Data(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.CalculationNode(), expected_db_statistics)

        new_db_statistics = self.backend.query_manager.get_creation_statistics()
        # I only check a few fields
        new_db_statistics = {k: v for k, v in new_db_statistics.items() if k in expected_db_statistics}

        expected_db_statistics = {
            k: dict(v) if isinstance(v, defaultdict) else v for k, v in expected_db_statistics.items()
        }

        self.assertEqual(new_db_statistics, expected_db_statistics)


class TestDoubleStar(AiidaTestCase):
    """
    In this test class we check if QueryBuilder returns the correct results
    when double star is provided as projection.
    """

    def test_statistics_default_class(self):

        # The expected result
        # pylint: disable=no-member
        expected_dict = {
            u'description': self.computer.description,
            u'scheduler_type': self.computer.get_scheduler_type(),
            u'hostname': self.computer.hostname,
            u'uuid': self.computer.uuid,
            u'name': self.computer.name,
            u'transport_type': self.computer.get_transport_type(),
            u'id': self.computer.id,
            u'metadata': self.computer.get_metadata(),
        }

        qb = orm.QueryBuilder()
        qb.append(orm.Computer, project=['**'])
        # We expect one result
        self.assertEqual(qb.count(), 1)

        # Get the one result record and check that the returned
        # data are correct
        res = list(qb.dict()[0].values())[0]
        self.assertDictEqual(res, expected_dict)

        # Ask the same query as above using queryhelp
        qh = {'project': {'computer': ['**']}, 'path': [{'tag': 'computer', 'cls': orm.Computer}]}
        qb = orm.QueryBuilder(**qh)
        # We expect one result
        self.assertEqual(qb.count(), 1)

        # Get the one result record and check that the returned
        # data are correct
        res = list(qb.dict()[0].values())[0]
        self.assertDictEqual(res, expected_dict)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that defines the exceptions thrown by AiiDA's export/import module.

Note: In order to not override the built-in `ImportError`, both `ImportError` and `ExportError` are prefixed with
    `Archive`.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from aiida.common.exceptions import AiidaException

__all__ = (
    'ExportImportException', 'ArchiveExportError', 'ArchiveImportError', 'CorruptArchive',
    'IncompatibleArchiveVersionError', 'ExportValidationError', 'ImportUniquenessError', 'ImportValidationError',
    'ArchiveMigrationError', 'MigrationValidationError', 'DanglingLinkError'
)


class ExportImportException(AiidaException):
    """Base class for all AiiDA export/import module exceptions."""


class ArchiveExportError(ExportImportException):
    """Base class for all AiiDA export exceptions."""


class ArchiveImportError(ExportImportException):
    """Base class for all AiiDA import exceptions."""


class CorruptArchive(ExportImportException):
    """Raised when an operation is applied to a corrupt export archive, e.g. missing files or invalid formats."""


class IncompatibleArchiveVersionError(ExportImportException):
    """Raised when trying to import an export archive with an incompatible schema version."""


class ExportValidationError(ArchiveExportError):
    """Raised when validation fails during export, e.g. for non-sealed ``ProcessNode`` s."""


class ImportUniquenessError(ArchiveImportError):
    """Raised when the user tries to violate a uniqueness constraint.

    Similar to :py:class:`~aiida.common.exceptions.UniquenessError`.
    """


class ImportValidationError(ArchiveImportError):
    """Raised when validation fails during import, e.g. for parameter types and values."""


class ArchiveMigrationError(ExportImportException):
    """Base class for all AiiDA export archive migration exceptions."""


class MigrationValidationError(ArchiveMigrationError):
    """Raised when validation fails during migration of export archives."""


class DanglingLinkError(MigrationValidationError):
    """Raised when an export archive is detected to contain dangling links when importing."""



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SQLA Log and LogCollection module"""
# pylint: disable=import-error,no-name-in-module
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from sqlalchemy.orm.exc import NoResultFound

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.sqlalchemy.models import log as models
from aiida.common import exceptions

from .. import BackendLog, BackendLogCollection
from . import entities
from . import utils


class SqlaLog(entities.SqlaModelEntity[models.DbLog], BackendLog):
    """SQLA Log backend entity"""

    MODEL_CLASS = models.DbLog

    def __init__(self, backend, time, loggername, levelname, dbnode_id, message='', metadata=None):
        # pylint: disable=too-many-arguments
        super(SqlaLog, self).__init__(backend)
        self._dbmodel = utils.ModelWrapper(
            models.DbLog(
                time=time,
                loggername=loggername,
                levelname=levelname,
                dbnode_id=dbnode_id,
                message=message,
                metadata=metadata
            )
        )

    @property
    def uuid(self):
        """
        Get the string representation of the UUID of the log entry
        """
        return str(self._dbmodel.uuid)

    @property
    def time(self):
        """
        Get the time corresponding to the entry
        """
        return self._dbmodel.time

    @property
    def loggername(self):
        """
        The name of the logger that created this entry
        """
        return self._dbmodel.loggername

    @property
    def levelname(self):
        """
        The name of the log level
        """
        return self._dbmodel.levelname

    @property
    def dbnode_id(self):
        """
        Get the id of the object that created the log entry
        """
        return self._dbmodel.dbnode_id

    @property
    def message(self):
        """
        Get the message corresponding to the entry
        """
        return self._dbmodel.message

    @property
    def metadata(self):
        """
        Get the metadata corresponding to the entry
        """
        return self._dbmodel._metadata  # pylint: disable=protected-access


class SqlaLogCollection(BackendLogCollection):
    """The SQLA collection for logs"""

    ENTITY_CLASS = SqlaLog

    def delete(self, log_id):
        """
        Remove a Log entry from the collection with the given id

        :param log_id: id of the Log to delete
        :type log_id: int

        :raises TypeError: if ``log_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Log with ID ``log_id`` is not found
        """
        if not isinstance(log_id, int):
            raise TypeError('log_id must be an int')

        session = get_scoped_session()

        try:
            session.query(models.DbLog).filter_by(id=log_id).one().delete()
            session.commit()
        except NoResultFound:
            session.rollback()
            raise exceptions.NotExistent("Log with id '{}' not found".format(log_id))

    def delete_all(self):
        """
        Delete all Log entries.

        :raises `~aiida.common.exceptions.IntegrityError`: if all Logs could not be deleted
        """
        session = get_scoped_session()

        try:
            session.query(models.DbLog).delete()
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exceptions.IntegrityError('Could not delete all Logs. Full exception: {}'.format(exc))

    def delete_many(self, filters):
        """
        Delete Logs based on ``filters``

        :param filters: similar to QueryBuilder filter
        :type filters: dict

        :return: (former) ``PK`` s of deleted Logs
        :rtype: list

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """
        from aiida.orm import Log, QueryBuilder

        # Checks
        if not isinstance(filters, dict):
            raise TypeError('filters must be a dictionary')
        if not filters:
            raise exceptions.ValidationError('filter must not be empty')

        # Apply filter and delete found entities
        builder = QueryBuilder().append(Log, filters=filters, project='id').all()
        entities_to_delete = [_[0] for _ in builder]
        for entity in entities_to_delete:
            self.delete(entity)

        # Return list of deleted entities' (former) PKs for checking
        return entities_to_delete



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Backend specific computer objects and methods"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import logging
import six

from . import backends

__all__ = ('BackendComputer', 'BackendComputerCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendComputer(backends.BackendEntity):
    """
    Base class to map a node in the DB + its permanent repository counterpart.

    Stores attributes starting with an underscore.

    Caches files and attributes before the first save, and saves everything only on store().
    After the call to store(), attributes cannot be changed.

    Only after storing (or upon loading from uuid) metadata can be modified
    and in this case they are directly set on the db.

    In the plugin, also set the _plugin_type_string, to be set in the DB in the 'type' field.
    """
    # pylint: disable=too-many-public-methods

    _logger = logging.getLogger(__name__)

    @abc.abstractproperty
    def is_stored(self):
        """
        Is the computer stored?

        :return: True if stored, False otherwise
        :rtype: bool
        """

    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractproperty
    def description(self):
        pass

    @abc.abstractproperty
    def hostname(self):
        pass

    @abc.abstractmethod
    def get_metadata(self):
        pass

    @abc.abstractmethod
    def set_metadata(self, metadata):
        """
        Set the metadata.

        .. note: You still need to call the .store() method to actually save
           data to the database! (The store method can be called multiple
           times, differently from AiiDA Node objects).
        """

    @abc.abstractmethod
    def get_name(self):
        pass

    @abc.abstractmethod
    def set_name(self, val):
        pass

    def get_hostname(self):
        """
        Get this computer hostname
        :rtype: str
        """

    @abc.abstractmethod
    def set_hostname(self, val):
        """
        Set the hostname of this computer
        :param val: The new hostname
        :type val: str
        """

    @abc.abstractmethod
    def get_description(self):
        pass

    @abc.abstractmethod
    def set_description(self, val):
        pass

    @abc.abstractmethod
    def get_scheduler_type(self):
        pass

    @abc.abstractmethod
    def set_scheduler_type(self, scheduler_type):
        pass

    @abc.abstractmethod
    def get_transport_type(self):
        pass

    @abc.abstractmethod
    def set_transport_type(self, transport_type):
        pass


@six.add_metaclass(abc.ABCMeta)
class BackendComputerCollection(backends.BackendCollection[BackendComputer]):
    """The collection of Computer entries."""

    ENTITY_CLASS = BackendComputer

    @abc.abstractmethod
    def delete(self, pk):
        """
        Delete an entry with the given pk

        :param pk: the pk of the entry to delete
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unit tests for the QueryBuilder ORM class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six

from aiida import orm
from aiida.backends.testbase import AiidaTestCase


class TestQueryBuilder(AiidaTestCase):
    """Unit tests for the QueryBuilder ORM class."""

    def test_first_multiple_projections(self):
        """Test `first()` returns correct types and numbers for multiple projections."""
        orm.Data().store()
        orm.Data().store()

        result = orm.QueryBuilder().append(orm.User, tag='user',
                                           project=['email']).append(orm.Data, with_user='user', project=['*']).first()

        self.assertEqual(type(result), list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], six.string_types)
        self.assertIsInstance(result[1], orm.Data)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Extras tests for the export and import routines"""
# pylint: disable=attribute-defined-outside-init
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import with_statement

import os

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.configuration import with_temp_dir
from aiida.tools.importexport import import_data, export


class TestAttributes(AiidaTestCase):
    """Test ex-/import cases related to Attributes"""

    def create_data_with_attr(self):
        self.data = orm.Data()
        self.data.label = 'my_test_data_node'
        self.data.set_attribute_many({'b': 2, 'c': 3})
        self.data.store()

    def import_attributes(self):
        """Import an AiiDA database"""
        import_data(self.export_file, silent=True)

        builder = orm.QueryBuilder().append(orm.Data, filters={'label': 'my_test_data_node'})

        self.assertEqual(builder.count(), 1)
        self.imported_node = builder.all()[0][0]

    @with_temp_dir
    def test_import_of_attributes(self, temp_dir):
        """Check if attributes are properly imported"""
        # Create Data with attributes
        self.create_data_with_attr()

        self.assertEqual(self.data.get_attribute('b'), 2)
        self.assertEqual(self.data.get_attribute('c'), 3)

        # Export
        self.export_file = os.path.join(temp_dir, 'export.aiida')
        export([self.data], outfile=self.export_file, silent=True)

        # Clean db
        self.reset_database()

        self.import_attributes()

        self.assertEqual(self.imported_node.get_attribute('b'), 2)
        self.assertEqual(self.imported_node.get_attribute('c'), 3)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions and classes to interact with AiiDA export archives."""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import io
import os
import sys
import tarfile
import zipfile

from wrapt import decorator

from aiida.common import json
from aiida.common.exceptions import ContentNotExistent, InvalidOperation
from aiida.common.folders import SandboxFolder
from aiida.tools.importexport.common.config import NODES_EXPORT_SUBFOLDER
from aiida.tools.importexport.common.exceptions import CorruptArchive

__all__ = ('Archive', 'extract_zip', 'extract_tar', 'extract_tree')


class Archive(object):
    """Utility class to operate on exported archive files or directories.

    The main usage should be to construct the class with the filepath of the export archive as an argument.
    The contents will be lazily unpacked into a sand box folder which is constructed upon entering the instance
    within a context and which will be automatically cleaned upon leaving that context. Example::

        with Archive('/some/path/archive.aiida') as archive:
            archive.version

    """

    FILENAME_DATA = 'data.json'
    FILENAME_METADATA = 'metadata.json'

    def __init__(self, filepath):
        self._filepath = filepath
        self._folder = None
        self._unpacked = False
        self._data = None
        self._meta_data = None

    def __enter__(self):
        """Instantiate a SandboxFolder into which the archive can be lazily unpacked."""
        self._folder = SandboxFolder()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Clean the sandbox folder if it was instatiated."""
        if self.folder:
            self.folder.erase()

    @decorator
    def ensure_within_context(wrapped, instance, args, kwargs):  # pylint: disable=no-self-argument
        """Decorator to ensure that the instance is called within a context manager."""
        if instance and not instance.folder:
            raise InvalidOperation('the Archive class should be used within a context')

        return wrapped(*args, **kwargs)  # pylint: disable=not-callable

    @decorator
    def ensure_unpacked(wrapped, instance, args, kwargs):  # pylint: disable=no-self-argument
        """Decorator to ensure that the archive is unpacked before entering the decorated function."""
        if instance and not instance.unpacked:
            instance.unpack()

        return wrapped(*args, **kwargs)  # pylint: disable=not-callable

    @ensure_within_context
    def unpack(self):
        """Unpack the archive and store the contents in a sandbox."""
        if os.path.isdir(self.filepath):
            extract_tree(self.filepath, self.folder)
        elif tarfile.is_tarfile(self.filepath):
            extract_tar(self.filepath, self.folder, silent=True, nodes_export_subfolder=NODES_EXPORT_SUBFOLDER)
        elif zipfile.is_zipfile(self.filepath):
            extract_zip(self.filepath, self.folder, silent=True, nodes_export_subfolder=NODES_EXPORT_SUBFOLDER)
        else:
            raise CorruptArchive('unrecognized archive format')

        if not self.folder.get_content_list():
            raise ContentNotExistent('the provided archive {} is empty'.format(self.filepath))

        self._unpacked = True

    @property
    def filepath(self):
        """Return the filepath of the archive

        :return: the archive filepath
        """
        return self._filepath

    @property
    def folder(self):
        """Return the sandbox folder

        :return: sandbox folder :class:`aiida.common.folders.SandboxFolder`
        """
        return self._folder

    @property
    @ensure_within_context
    def data(self):
        """Return the loaded content of the data file

        :return: dictionary with contents of data file
        """
        if self._data is None:
            self._data = self._read_json_file(self.FILENAME_DATA)

        return self._data

    @property
    @ensure_within_context
    def meta_data(self):
        """Return the loaded content of the meta data file

        :return: dictionary with contents of meta data file
        """
        if self._meta_data is None:
            self._meta_data = self._read_json_file(self.FILENAME_METADATA)

        return self._meta_data

    @property
    @ensure_within_context
    def unpacked(self):
        """Return whether the archive has been unpacked into the sandbox folder."""
        return self._unpacked

    @ensure_within_context
    def get_info(self):
        """Return a dictionary with basic information about the archive.

        :return: a dictionary with basic details
        """
        info = {
            'version aiida': self.version_aiida,
            'version format': self.version_format,
        }

        if self.conversion_info:
            info['conversion info'] = '\n'.join(self.conversion_info)

        return info

    @ensure_within_context
    def get_data_statistics(self):
        """Return dictionary with statistics about data content, i.e. how many entries of each entity type it contains.

        :return: a dictionary with basic details
        """
        export_data = self.data.get('export_data', {})
        links_data = self.data.get('links_uuid', {})

        computers = export_data.get('Computer', {})
        groups = export_data.get('Group', {})
        nodes = export_data.get('Node', {})
        users = export_data.get('User', {})

        statistics = {
            'computers': len(computers),
            'groups': len(groups),
            'links': len(links_data),
            'nodes': len(nodes),
            'users': len(users),
        }

        return statistics

    @property
    @ensure_within_context
    def version_aiida(self):
        """Return the version of AiiDA the archive was created with.

        :return: version number
        """
        return self.meta_data['aiida_version']

    @property
    @ensure_within_context
    def version_format(self):
        """Return the version of the archive format.

        :return: version number
        """
        return self.meta_data['export_version']

    @property
    @ensure_within_context
    def conversion_info(self):
        """Return information about migration events that were applied to this archive.

        :return: list of conversion notifications
        """
        try:
            return self.meta_data['conversion_info']
        except KeyError:
            return None

    @ensure_within_context
    @ensure_unpacked
    def _read_json_file(self, filename):
        """Read the contents of a JSON file from the unpacked archive contents.

        :param filename: the filename relative to the sandbox folder
        :return: a dictionary with the loaded JSON content
        """
        with io.open(self.folder.get_abs_path(filename), 'r', encoding='utf8') as fhandle:
            return json.load(fhandle)


def extract_zip(infile, folder, nodes_export_subfolder=None, silent=False):
    """
    Extract the nodes to be imported from a zip file.

    :param infile: file path
    :type infile: str

    :param folder: a temporary folder used to extract the file tree
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`

    :param nodes_export_subfolder: name of the subfolder for AiiDA nodes
    :type nodes_export_subfolder: str

    :param silent: suppress debug print
    :type silent: bool

    :raises TypeError: if parameter types are not respected
    :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: if the archive misses files or files have
        incorrect formats
    """
    # pylint: disable=fixme
    if not silent:
        print('READING DATA AND METADATA...')

    if nodes_export_subfolder:
        if not isinstance(nodes_export_subfolder, str):
            raise TypeError('nodes_export_subfolder must be a string')
    else:
        nodes_export_subfolder = NODES_EXPORT_SUBFOLDER

    try:
        with zipfile.ZipFile(infile, 'r', allowZip64=True) as handle:

            if not handle.namelist():
                raise CorruptArchive('no files detected')

            try:
                handle.extract(path=folder.abspath, member='metadata.json')
            except KeyError:
                raise CorruptArchive('required file `metadata.json` is not included')

            try:
                handle.extract(path=folder.abspath, member='data.json')
            except KeyError:
                raise CorruptArchive('required file `data.json` is not included')

            if not silent:
                print('EXTRACTING NODE DATA...')

            for membername in handle.namelist():
                # Check that we are only exporting nodes within the subfolder!
                # TODO: better check such that there are no .. in the
                # path; use probably the folder limit checks
                if not membername.startswith(nodes_export_subfolder + os.sep):
                    continue
                handle.extract(path=folder.abspath, member=membername)
    except zipfile.BadZipfile:
        raise ValueError('The input file format for import is not valid (not a zip file)')


def extract_tar(infile, folder, nodes_export_subfolder=None, silent=False):
    """
    Extract the nodes to be imported from a (possibly zipped) tar file.

    :param infile: file path
    :type infile: str

    :param folder: a temporary fodler used to extract the file tree
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`

    :param nodes_export_subfolder: name of the subfolder for AiiDA nodes
    :type nodes_export_subfolder: str

    :param silent: suppress debug print
    :type silent: bool

    :raises TypeError: if parameter types are not respected
    :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: if the archive misses files or files have
        incorrect formats
    """
    # pylint: disable=fixme
    if not silent:
        print('READING DATA AND METADATA...')

    if nodes_export_subfolder:
        if not isinstance(nodes_export_subfolder, str):
            raise TypeError('nodes_export_subfolder must be a string')
    else:
        nodes_export_subfolder = NODES_EXPORT_SUBFOLDER

    try:
        with tarfile.open(infile, 'r:*', format=tarfile.PAX_FORMAT) as handle:

            try:
                handle.extract(path=folder.abspath, member=handle.getmember('metadata.json'))
            except KeyError:
                raise CorruptArchive('required file `metadata.json` is not included')

            try:
                handle.extract(path=folder.abspath, member=handle.getmember('data.json'))
            except KeyError:
                raise CorruptArchive('required file `data.json` is not included')

            if not silent:
                print('EXTRACTING NODE DATA...')

            for member in handle.getmembers():
                if member.isdev():
                    # safety: skip if character device, block device or FIFO
                    print('WARNING, device found inside the import file: {}'.format(member.name), file=sys.stderr)
                    continue
                if member.issym() or member.islnk():
                    # safety: in export, I set dereference=True therefore
                    # there should be no symbolic or hard links.
                    print('WARNING, link found inside the import file: {}'.format(member.name), file=sys.stderr)
                    continue
                # Check that we are only exporting nodes within the subfolder!
                # TODO: better check such that there are no .. in the
                # path; use probably the folder limit checks
                if not member.name.startswith(nodes_export_subfolder + os.sep):
                    continue
                handle.extract(path=folder.abspath, member=member)
    except tarfile.ReadError:
        raise ValueError('The input file format for import is not valid (1)')


def extract_tree(infile, folder):
    """
    Prepare to import nodes from plain file system tree.

    :param infile: path
    :type infile: str

    :param folder: a temporary folder used to extract the file tree
    :type folder: :py:class:`~aiida.common.folders.SandboxFolder`
    """

    def add_files(args, path, files):
        """Add files to a folder."""
        folder = args['folder']
        root = args['root']
        for filename in files:
            fullpath = os.path.join(path, filename)
            relpath = os.path.relpath(fullpath, root)
            if os.path.isdir(fullpath):
                if os.path.dirname(relpath) != '':
                    folder.get_subfolder(os.path.dirname(relpath) + os.sep, create=True)
            elif not os.path.isfile(fullpath):
                continue
            if os.path.dirname(relpath) != '':
                folder.get_subfolder(os.path.dirname(relpath) + os.sep, create=True)
            folder.insert_path(os.path.abspath(fullpath), relpath)

    os.walk(infile, add_files, {'folder': folder, 'root': infile})



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for testing components from the workflow engine"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import plumpy

from aiida.orm import Data, WorkflowNode
from aiida.engine import Process


class DummyProcess(Process):
    """A Process that does nothing when it runs."""

    _node_class = WorkflowNode

    @classmethod
    def define(cls, spec):
        super(DummyProcess, cls).define(spec)
        spec.inputs.valid_type = Data
        spec.outputs.valid_type = Data

    def run(self):
        pass


class AddProcess(Process):
    """A simple Process that adds two integers."""

    _node_class = WorkflowNode

    @classmethod
    def define(cls, spec):
        super(AddProcess, cls).define(spec)
        spec.input('a', required=True)
        spec.input('b', required=True)
        spec.output('result', required=True)

    def run(self):
        summed = self.inputs.a + self.inputs.b
        self.out(summed.store())


class BadOutput(Process):
    """A Process that emits an output that isn't an AiiDA Data type."""

    _node_class = WorkflowNode

    @classmethod
    def define(cls, spec):
        super(BadOutput, cls).define(spec)
        spec.outputs.valid_type = Data

    def run(self):
        self.out('bad_output', 5)


class ExceptionProcess(Process):
    """A Process that raises a RuntimeError when run."""

    _node_class = WorkflowNode

    def run(self):  # pylint: disable=no-self-use
        raise RuntimeError('CRASH')


class WaitProcess(Process):
    """A Process that waits until it is asked to continue."""

    _node_class = WorkflowNode

    def run(self):
        return plumpy.Wait(self.next_step)

    def next_step(self):
        pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member,import-error,no-name-in-module
"""This migration cleans the log records from non-Node entity records.
It removes from the DbLog table the legacy workflow records and records
that correspond to an unknown entity and places them to corresponding files.

This migration corresponds to the 0024_dblog_update Django migration.

Revision ID: 041a79fc615f
Revises: 7ca08c391c49
Create Date: 2018-12-28 15:53:14.596810
"""
# pylint: disable=wrong-import-order
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import sys
from six.moves import zip
import click
import sqlalchemy as sa
from sqlalchemy.sql import text
from alembic import op

from aiida.backends.general.migrations.utils import dumps_json
from aiida.manage import configuration

# revision identifiers, used by Alembic.
revision = '041a79fc615f'
down_revision = '7ca08c391c49'
branch_labels = None
depends_on = None

# The values that will be exported for the log records that will be deleted
values_to_export = ['id', 'time', 'loggername', 'levelname', 'objpk', 'objname', 'message', 'metadata']


def get_legacy_workflow_log_number(connection):
    """ Get the number of the log records that correspond to legacy workflows """
    return connection.execute(
        text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'aiida.workflows.user.%')
        """
        )
    ).fetchall()[0][0]


def get_unknown_entity_log_number(connection):
    """ Get the number of the log records that correspond to unknown entities """
    return connection.execute(
        text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname NOT LIKE 'node.%') AND
            (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
        """
        )
    ).fetchall()[0][0]


def get_logs_with_no_nodes_number(connection):
    """ Get the number of the log records that correspond to nodes that were deleted """
    return connection.execute(
        text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
        """
        )
    ).fetchall()[0][0]


def get_serialized_legacy_workflow_logs(connection):
    """ Get the serialized log records that correspond to legacy workflows """
    query = connection.execute(
        text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'aiida.workflows.user.%')
        """
        )
    )
    res = list()
    for row in query:
        res.append(dict(list(zip(row.keys(), row))))
    return dumps_json(res)


def get_serialized_unknown_entity_logs(connection):
    """ Get the serialized log records that correspond to unknown entities """
    query = connection.execute(
        text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname NOT LIKE 'node.%') AND
            (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
        """
        )
    )
    res = list()
    for row in query:
        res.append(dict(list(zip(row.keys(), row))))
    return dumps_json(res)


def get_serialized_logs_with_no_nodes(connection):
    """ Get the serialized log records that correspond to nodes that were deleted """
    query = connection.execute(
        text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
        """
        )
    )
    res = list()
    for row in query:
        res.append(dict(list(zip(row.keys(), row))))
    return dumps_json(res)


def export_and_clean_workflow_logs(connection):
    """
    Export the logs records that correspond to legacy workflows and to unknown entities (place them to files
    and remove them from the DbLog table).
    """
    from tempfile import NamedTemporaryFile

    lwf_no_number = get_legacy_workflow_log_number(connection)
    other_number = get_unknown_entity_log_number(connection)
    log_no_node_number = get_logs_with_no_nodes_number(connection)

    # If there are no legacy workflow log records or log records of an unknown entity
    if lwf_no_number == 0 and other_number == 0 and log_no_node_number == 0:
        return

    if not configuration.PROFILE.is_test_profile:
        click.echo(
            'We found {} log records that correspond to legacy workflows and {} log records to correspond '
            'to an unknown entity.'.format(lwf_no_number, other_number)
        )
        click.echo(
            'These records will be removed from the database and exported to JSON files to the current directory).'
        )
        proceed = click.confirm('Would you like to proceed?', default=True)
        if not proceed:
            sys.exit(1)

    delete_on_close = configuration.PROFILE.is_test_profile

    # Exporting the legacy workflow log records
    if lwf_no_number != 0:

        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='legagy_wf_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_legacy_workflow_logs(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                click.echo('Exported legacy workflow logs to {}'.format(filename))

        # Now delete the records
        connection.execute(
            text(
                """
            DELETE FROM db_dblog
            WHERE
                (db_dblog.objname LIKE 'aiida.workflows.user.%')
            """
            )
        )

    # Exporting unknown log records
    if other_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='unknown_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_unknown_entity_logs(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                click.echo('Exported unexpected entity logs to {}'.format(filename))

        # Now delete the records
        connection.execute(
            text(
                """
            DELETE FROM db_dblog WHERE
                (db_dblog.objname NOT LIKE 'node.%') AND
                (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
            """
            )
        )

    # Exporting log records that don't correspond to nodes
    if log_no_node_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='no_node_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_logs_with_no_nodes(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                click.echo('Exported entity logs that don\'t correspond to nodes to {}'.format(filename))

        # Now delete the records
        connection.execute(
            text(
                """
            DELETE FROM db_dblog WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
            """
            )
        )


def upgrade():
    """
    Changing the log table columns to use uuid to reference remote objects and log entries.
    Upgrade function.
    """
    connection = op.get_bind()

    # Clean data
    export_and_clean_workflow_logs(connection)

    # Create the dbnode_id column and add the necessary index
    op.add_column('db_dblog', sa.Column('dbnode_id', sa.INTEGER(), autoincrement=False, nullable=True))
    # Transfer data to dbnode_id from objpk
    connection.execute(text("""UPDATE db_dblog SET dbnode_id=objpk"""))

    op.create_foreign_key(
        None,
        'db_dblog',
        'db_dbnode', ['dbnode_id'], ['id'],
        ondelete=u'CASCADE',
        initially=u'DEFERRED',
        deferrable=True
    )

    # Update the dbnode_id column to not nullable
    op.alter_column('db_dblog', 'dbnode_id', nullable=False)

    # Remove the objpk column
    op.drop_column('db_dblog', 'objpk')

    # Remove the objname column
    op.drop_column('db_dblog', 'objname')

    # Remove objpk and objname from metadata dictionary
    connection.execute(text("""UPDATE db_dblog SET metadata = metadata - 'objpk' - 'objname' """))


def downgrade():
    """
    Downgrade function to the previous schema.
    """
    # Create an empty column objname (the data is permanently lost)
    op.add_column('db_dblog', sa.Column('objname', sa.VARCHAR(length=255), autoincrement=False, nullable=True))
    op.create_index('ix_db_dblog_objname', 'db_dblog', ['objname'])

    # Creating a column objpk -
    op.add_column('db_dblog', sa.Column('objpk', sa.INTEGER(), autoincrement=False, nullable=True))

    # Copy the data back to objpk from dbnode_id
    op.execute(text("""UPDATE db_dblog SET objpk=dbnode_id"""))

    # Removing the column dbnode_id
    op.drop_column('db_dblog', 'dbnode_id')

    # Populate objname with correct values
    op.execute(
        text("""UPDATE db_dblog SET objname=db_dbnode.type
    FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk""")
    )

    # Enrich metadata with objpk and objname if these keys don't exist
    op.execute(
        text(
            """UPDATE db_dblog SET metadata = jsonb_set(metadata, '{"objpk"}', to_jsonb(objpk))
    WHERE NOT (metadata ?| '{"objpk"}') """
        )
    )
    op.execute(
        text(
            """UPDATE db_dblog SET metadata = jsonb_set(metadata, '{"objname"}', to_jsonb(objname))
    WHERE NOT (metadata ?| '{"objname"}') """
        )
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module,invalid-name
"""
Tests for the migrations of the attributes, extras and settings from EAV to JSONB
Migration 0037_attributes_extras_settings_json
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from aiida.backends.djsite.db.subtests.migrations.test_migrations_common import TestMigrations
from aiida.backends.general.migrations.calc_state import STATE_MAPPING


class TestLegacyJobCalcStateDataMigration(TestMigrations):
    """Test the migration that performs a data migration of legacy `JobCalcState`."""

    migrate_from = '0037_attributes_extras_settings_json'
    migrate_to = '0038_data_migration_legacy_job_calculations'

    def setUpBeforeMigration(self):
        self.nodes = {}

        for state in STATE_MAPPING:
            node = self.DbNode(
                node_type='process.calculation.calcjob.CalcJobNode.',
                user_id=self.default_user.id,
                attributes={'state': state}
            )
            node.save()

            self.nodes[state] = node.id

    def test_data_migrated(self):
        """Verify that the `process_state`, `process_status` and `exit_status` are set correctly."""
        for state, pk in self.nodes.items():
            node = self.load_node(pk)
            self.assertEqual(node.attributes.get('process_state', None), STATE_MAPPING[state].process_state)
            self.assertEqual(node.attributes.get('process_status', None), STATE_MAPPING[state].process_status)
            self.assertEqual(node.attributes.get('exit_status', None), STATE_MAPPING[state].exit_status)
            self.assertEqual(
                node.attributes.get('process_label'), 'Legacy JobCalculation'
            )  # All nodes should have this label
            self.assertIsNone(node.attributes.get('state', None))  # The old state should have been removed

            exit_status = node.attributes.get('exit_status', None)
            if exit_status is not None:
                self.assertIsInstance(exit_status, six.integer_types)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Click parameter type for AiiDA Plugins."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six
import click

from aiida.cmdline.utils import decorators
from aiida.common import exceptions
from aiida.plugins.entry_point import ENTRY_POINT_STRING_SEPARATOR, ENTRY_POINT_GROUP_PREFIX, EntryPointFormat
from aiida.plugins.entry_point import format_entry_point_string, get_entry_point_string_format
from aiida.plugins.entry_point import get_entry_point, get_entry_points, get_entry_point_groups


class PluginParamType(click.ParamType):
    """
    AiiDA Plugin name parameter type.

    :param group: string or tuple of strings, where each is a valid entry point group. Adding the `aiida.`
        prefix is optional. If it is not detected it will be prepended internally.
    :param load: when set to True, convert will not return the entry point, but the loaded entry point

    Usage::

        click.option(... type=PluginParamType(group='aiida.calculations')

    or::

        click.option(... type=PluginParamType(group=('calculations', 'data'))

    """
    name = 'plugin'

    def __init__(self, group=None, load=False, *args, **kwargs):
        """
        Validate that group is either a string or a tuple of valid entry point groups, or if it
        is not specified use the tuple of all recognized entry point groups.
        """
        # pylint: disable=keyword-arg-before-vararg
        valid_entry_point_groups = get_entry_point_groups()

        if group is None:
            self._groups = tuple(valid_entry_point_groups)
        else:
            if isinstance(group, six.string_types):
                invalidated_groups = tuple([group])
            elif isinstance(group, tuple):
                invalidated_groups = group
            else:
                raise ValueError('invalid type for group')

            groups = []

            for grp in invalidated_groups:

                if not grp.startswith(ENTRY_POINT_GROUP_PREFIX):
                    grp = ENTRY_POINT_GROUP_PREFIX + grp

                if grp not in valid_entry_point_groups:
                    raise ValueError('entry point group {} is not recognized'.format(grp))

                groups.append(grp)

            self._groups = tuple(groups)

        self._init_entry_points()
        self.load = load

        super(PluginParamType, self).__init__(*args, **kwargs)

    def _init_entry_points(self):
        """
        Populate entry point information that will be used later on.  This should only be called
        once in the constructor after setting self.groups because the groups should not be changed
        after instantiation
        """
        self._entry_points = [(group, entry_point) for group in self.groups for entry_point in get_entry_points(group)]
        self._entry_point_names = [entry_point.name for group in self.groups for entry_point in get_entry_points(group)]

    @property
    def groups(self):
        return self._groups

    @property
    def has_potential_ambiguity(self):
        """
        Returns whether the set of supported entry point groups can lead to ambiguity when only an entry point name
        is specified. This will happen if one ore more groups share an entry point with a common name
        """
        return len(self._entry_point_names) != len(set(self._entry_point_names))

    def get_valid_arguments(self):
        """
        Return a list of all available plugins for the groups configured for this PluginParamType instance.
        If the entry point names are not unique, because there are multiple groups that contain an entry
        point that has an identical name, we need to prefix the names with the full group name

        :returns: list of valid entry point strings
        """
        if self.has_potential_ambiguity:
            fmt = EntryPointFormat.FULL
            return sorted([format_entry_point_string(group, ep.name, fmt=fmt) for group, ep in self._entry_points])

        return sorted(self._entry_point_names)

    def get_possibilities(self, incomplete=''):
        """
        Return a list of plugins starting with incomplete
        """
        if incomplete == '':
            return self.get_valid_arguments()

        # If there is a chance of ambiguity we always return the entry point string in FULL format, otherwise
        # return the possibilities in the same format as the incomplete. Note that this may have some unexpected
        # effects. For example if incomplete equals `aiida.` or `calculations` it will be detected as the MINIMAL
        # format, even though they would also be the valid beginnings of a FULL or PARTIAL format, except that we
        # cannot know that for sure at this time
        if self.has_potential_ambiguity:
            possibilites = [eps for eps in self.get_valid_arguments() if eps.startswith(incomplete)]
        else:
            possibilites = []
            fmt = get_entry_point_string_format(incomplete)

            for group, entry_point in self._entry_points:
                entry_point_string = format_entry_point_string(group, entry_point.name, fmt=fmt)
                if entry_point_string.startswith(incomplete):
                    possibilites.append(entry_point_string)

        return possibilites

    def complete(self, ctx, incomplete):  # pylint: disable=unused-argument
        """
        Return possible completions based on an incomplete value

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        return [(p, '') for p in self.get_possibilities(incomplete=incomplete)]

    def get_missing_message(self, param):
        return 'Possible arguments are:\n\n' + '\n'.join(self.get_valid_arguments())

    def get_entry_point_from_string(self, entry_point_string):
        """
        Validate a given entry point string, which means that it should have a valid entry point string format
        and that the entry point unambiguously corresponds to an entry point in the groups configured for this
        instance of PluginParameterType.

        :returns: the entry point if valid
        :raises: ValueError if the entry point string is invalid
        """
        group = None
        name = None

        entry_point_format = get_entry_point_string_format(entry_point_string)

        if entry_point_format in (EntryPointFormat.FULL, EntryPointFormat.PARTIAL):

            group, name = entry_point_string.split(ENTRY_POINT_STRING_SEPARATOR)

            if entry_point_format == EntryPointFormat.PARTIAL:
                group = ENTRY_POINT_GROUP_PREFIX + group

            if group not in self.groups:
                raise ValueError('entry point group {} is not supported by this parameter')

        elif entry_point_format == EntryPointFormat.MINIMAL:

            name = entry_point_string
            matching_groups = [group for group, entry_point in self._entry_points if entry_point.name == name]

            if len(matching_groups) > 1:
                raise ValueError(
                    "entry point '{}' matches more than one valid entry point group [{}], "
                    'please specify an explicit group prefix'.format(name, ' '.join(matching_groups))
                )
            elif not matching_groups:
                raise ValueError(
                    "entry point '{}' is not valid for any of the allowed "
                    'entry point groups: {}'.format(name, ' '.join(self.groups))
                )
            else:
                group = matching_groups[0]

        else:
            ValueError('invalid entry point string format: {}'.format(entry_point_string))

        try:
            entry_point = get_entry_point(group, name)
        except exceptions.EntryPointError as exception:
            raise ValueError(exception)

        return entry_point

    @decorators.with_dbenv()
    def convert(self, value, param, ctx):
        """
        Convert the string value to an entry point instance, if the value can be successfully parsed
        into an actual entry point. Will raise click.BadParameter if validation fails.
        """
        if not value:
            raise click.BadParameter('plugin name cannot be empty')

        try:
            entry_point = self.get_entry_point_from_string(value)
        except ValueError as exception:
            raise click.BadParameter(str(exception))

        if self.load:
            try:
                return entry_point.load()
            except exceptions.LoadingEntryPointError as exception:
                raise click.BadParameter(str(exception))
        else:
            return entry_point



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import functools
import logging
import sys
import tempfile

import six
from tornado.gen import coroutine, Return

import plumpy

from aiida.common.datastructures import CalcJobState
from aiida.common.exceptions import TransportTaskException
from aiida.engine.daemon import execmanager
from aiida.engine.utils import exponential_backoff_retry, interruptable_task
from aiida.schedulers.datastructures import JobState

from ..process import ProcessState

UPLOAD_COMMAND = 'upload'
SUBMIT_COMMAND = 'submit'
UPDATE_COMMAND = 'update'
RETRIEVE_COMMAND = 'retrieve'
KILL_COMMAND = 'kill'

TRANSPORT_TASK_RETRY_INITIAL_INTERVAL = 20
TRANSPORT_TASK_MAXIMUM_ATTEMTPS = 5

logger = logging.getLogger(__name__)


@coroutine
def task_upload_job(node, transport_queue, calc_info, script_filename, cancellable):
    """
    Transport task that will attempt to upload the files of a job calculation to the remote

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param calc_info: the calculation info datastructure returned by `CalcJobNode._presubmit`
    :param script_filename: the job launch script returned by `CalcJobNode._presubmit`
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.engine.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    if node.get_state() == CalcJobState.SUBMITTING:
        logger.warning('CalcJob<{}> already marked as SUBMITTING, skipping task_update_job'.format(node.pk))
        raise Return(True)

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.computer.get_authinfo(node.user)

    @coroutine
    def do_upload():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)
            raise Return(execmanager.upload_calculation(node, transport, calc_info, script_filename))

    try:
        logger.info('scheduled request to upload CalcJob<{}>'.format(node.pk))
        result = yield exponential_backoff_retry(
            do_upload, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.CancelledError)
    except plumpy.CancelledError:
        pass
    except Exception:
        logger.warning('uploading CalcJob<{}> failed'.format(node.pk))
        raise TransportTaskException('upload_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('uploading CalcJob<{}> successful'.format(node.pk))
        node.set_state(CalcJobState.SUBMITTING)
        raise Return(result)


@coroutine
def task_submit_job(node, transport_queue, calc_info, script_filename, cancellable):
    """
    Transport task that will attempt to submit a job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param calc_info: the calculation info datastructure returned by `CalcJobNode._presubmit`
    :param script_filename: the job launch script returned by `CalcJobNode._presubmit`
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.engine.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    if node.get_state() == CalcJobState.WITHSCHEDULER:
        assert node.get_job_id() is not None, 'job is WITHSCHEDULER, however, it does not have a job id'
        logger.warning('CalcJob<{}> already marked as WITHSCHEDULER, skipping task_submit_job'.format(node.pk))
        raise Return(node.get_job_id())

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.computer.get_authinfo(node.user)

    @coroutine
    def do_submit():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)
            raise Return(execmanager.submit_calculation(node, transport, calc_info, script_filename))

    try:
        logger.info('scheduled request to submit CalcJob<{}>'.format(node.pk))
        result = yield exponential_backoff_retry(
            do_submit, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        pass
    except Exception:
        logger.warning('submitting CalcJob<{}> failed'.format(node.pk))
        raise TransportTaskException('submit_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('submitting CalcJob<{}> successful'.format(node.pk))
        node.set_state(CalcJobState.WITHSCHEDULER)
        raise Return(result)


@coroutine
def task_update_job(node, job_manager, cancellable):
    """
    Transport task that will attempt to update the scheduler status of the job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :type node: :class:`aiida.orm.nodes.process.calculation.calcjob.CalcJobNode`
    :param job_manager: The job manager
    :type job_manager: :class:`aiida.engine.processes.calcjobs.manager.JobManager`
    :param cancellable: A cancel flag
    :type cancellable: :class:`aiida.engine.utils.InterruptableFuture`
    :raises: Return containing True if the tasks was successfully completed, False otherwise
    """
    if node.get_state() == CalcJobState.RETRIEVING:
        logger.warning('CalcJob<{}> already marked as RETRIEVING, skipping task_update_job'.format(node.pk))
        raise Return(True)

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.computer.get_authinfo(node.user)
    job_id = node.get_job_id()

    @coroutine
    def do_update():
        # Get the update request
        with job_manager.request_job_info_update(authinfo, job_id) as update_request:
            job_info = yield cancellable.with_interrupt(update_request)

        if job_info is None:
            # If the job is computed or not found assume it's done
            node.set_scheduler_state(JobState.DONE)
            job_done = True
        else:
            node.set_last_job_info(job_info)
            node.set_scheduler_state(job_info.job_state)
            job_done = job_info.job_state == JobState.DONE

        raise Return(job_done)

    try:
        logger.info('scheduled request to update CalcJob<{}>'.format(node.pk))
        job_done = yield exponential_backoff_retry(
            do_update, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('updating CalcJob<{}> failed'.format(node.pk))
        raise TransportTaskException('update_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('updating CalcJob<{}> successful'.format(node.pk))
        if job_done:
            node.set_state(CalcJobState.RETRIEVING)

        raise Return(job_done)


@coroutine
def task_retrieve_job(node, transport_queue, retrieved_temporary_folder, cancellable):
    """
    Transport task that will attempt to retrieve all files of a completed job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.engine.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    if node.get_state() == CalcJobState.PARSING:
        logger.warning('CalcJob<{}> already marked as PARSING, skipping task_retrieve_job'.format(node.pk))
        raise Return(True)

    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    authinfo = node.computer.get_authinfo(node.user)

    @coroutine
    def do_retrieve():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)
            raise Return(execmanager.retrieve_calculation(node, transport, retrieved_temporary_folder))

    try:
        logger.info('scheduled request to retrieve CalcJob<{}>'.format(node.pk))
        result = yield exponential_backoff_retry(
            do_retrieve, initial_interval, max_attempts, logger=node.logger, ignore_exceptions=plumpy.Interruption)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('retrieving CalcJob<{}> failed'.format(node.pk))
        raise TransportTaskException('retrieve_calculation failed {} times consecutively'.format(max_attempts))
    else:
        node.set_state(CalcJobState.PARSING)
        logger.info('retrieving CalcJob<{}> successful'.format(node.pk))
        raise Return(result)


@coroutine
def task_kill_job(node, transport_queue, cancellable):
    """
    Transport task that will attempt to kill a job calculation

    The task will first request a transport from the queue. Once the transport is yielded, the relevant execmanager
    function is called, wrapped in the exponential_backoff_retry coroutine, which, in case of a caught exception, will
    retry after an interval that increases exponentially with the number of retries, for a maximum number of retries.
    If all retries fail, the task will raise a TransportTaskException

    :param node: the node that represents the job calculation
    :param transport_queue: the TransportQueue from which to request a Transport
    :param cancellable: the cancelled flag that will be queried to determine whether the task was cancelled
    :type cancellable: :class:`aiida.engine.utils.InterruptableFuture`
    :raises: Return if the tasks was successfully completed
    :raises: TransportTaskException if after the maximum number of retries the transport task still excepted
    """
    initial_interval = TRANSPORT_TASK_RETRY_INITIAL_INTERVAL
    max_attempts = TRANSPORT_TASK_MAXIMUM_ATTEMTPS

    if node.get_state() in [CalcJobState.UPLOADING, CalcJobState.SUBMITTING]:
        logger.warning('CalcJob<{}> killed, it was in the {} state'.format(node.pk, node.get_state()))
        raise Return(True)

    authinfo = node.computer.get_authinfo(node.user)

    @coroutine
    def do_kill():
        with transport_queue.request_transport(authinfo) as request:
            transport = yield cancellable.with_interrupt(request)
            raise Return(execmanager.kill_calculation(node, transport))

    try:
        logger.info('scheduled request to kill CalcJob<{}>'.format(node.pk))
        result = yield exponential_backoff_retry(do_kill, initial_interval, max_attempts, logger=node.logger)
    except plumpy.Interruption:
        raise
    except Exception:
        logger.warning('killing CalcJob<{}> failed'.format(node.pk))
        raise TransportTaskException('kill_calculation failed {} times consecutively'.format(max_attempts))
    else:
        logger.info('killing CalcJob<{}> successful'.format(node.pk))
        node.set_scheduler_state(JobState.DONE)
        raise Return(result)


class Waiting(plumpy.Waiting):
    """The waiting state for the `CalcJob` process."""

    def __init__(self, process, done_callback, msg=None, data=None):
        super(Waiting, self).__init__(process, done_callback, msg, data)
        self._task = None
        self._killing = None

    def load_instance_state(self, saved_state, load_context):
        super(Waiting, self).load_instance_state(saved_state, load_context)
        self._task = None
        self._killing = None

    @coroutine
    def execute(self):

        node = self.process.node
        transport_queue = self.process.runner.transport

        if isinstance(self.data, tuple):
            command = self.data[0]
            args = self.data[1:]
        else:
            command = self.data

        process_status = 'Waiting for transport task: {}'.format(command)

        try:

            if command == UPLOAD_COMMAND:
                node.set_process_status(process_status)
                calc_info, script_filename = yield self._launch_task(task_upload_job, node, transport_queue, *args)
                raise Return(self.submit(calc_info, script_filename))

            elif command == SUBMIT_COMMAND:
                node.set_process_status(process_status)
                yield self._launch_task(task_submit_job, node, transport_queue, *args)
                raise Return(self.update())

            elif self.data == UPDATE_COMMAND:
                job_done = False

                while not job_done:
                    scheduler_state = node.get_scheduler_state()
                    scheduler_state_string = scheduler_state.name if scheduler_state else 'UNKNOWN'
                    process_status = 'Monitoring scheduler: job state {}'.format(scheduler_state_string)
                    node.set_process_status(process_status)
                    job_done = yield self._launch_task(task_update_job, node, self.process.runner.job_manager)

                raise Return(self.retrieve())

            elif self.data == RETRIEVE_COMMAND:
                node.set_process_status(process_status)
                # Create a temporary folder that has to be deleted by JobProcess.retrieved after successful parsing
                temp_folder = tempfile.mkdtemp()
                yield self._launch_task(task_retrieve_job, node, transport_queue, temp_folder)
                raise Return(self.parse(temp_folder))

            else:
                raise RuntimeError('Unknown waiting command')

        except TransportTaskException as exception:
            raise plumpy.PauseInterruption('Pausing after failed transport task: {}'.format(exception))
        except plumpy.KillInterruption:
            exc_info = sys.exc_info()
            yield self._launch_task(task_kill_job, node, transport_queue)
            self._killing.set_result(True)
            six.reraise(*exc_info)
        except Return:
            node.set_process_status(None)
            raise
        except (plumpy.Interruption, plumpy.CancelledError):
            node.set_process_status('Transport task {} was interrupted'.format(command))
            raise
        finally:
            # If we were trying to kill but we didn't deal with it, make sure it's set here
            if self._killing and not self._killing.done():
                self._killing.set_result(False)

    @coroutine
    def _launch_task(self, coro, *args, **kwargs):
        task_fn = functools.partial(coro, *args, **kwargs)
        try:
            self._task = interruptable_task(task_fn)
            result = yield self._task
            raise Return(result)
        finally:
            self._task = None

    def upload(self, calc_info, script_filename):
        """Return the `Waiting` state that will `upload` the `CalcJob`."""
        return self.create_state(
            ProcessState.WAITING,
            None,
            msg='Waiting for calculation folder upload',
            data=(UPLOAD_COMMAND, calc_info, script_filename))

    def submit(self, calc_info, script_filename):
        """Return the `Waiting` state that will `submit` the `CalcJob`."""
        return self.create_state(
            ProcessState.WAITING,
            None,
            msg='Waiting for scheduler submission',
            data=(SUBMIT_COMMAND, calc_info, script_filename))

    def update(self):
        """Return the `Waiting` state that will `update` the `CalcJob`."""
        return self.create_state(
            ProcessState.WAITING, None, msg='Waiting for scheduler update', data=UPDATE_COMMAND)

    def retrieve(self):
        """Return the `Waiting` state that will `retrieve` the `CalcJob`."""
        return self.create_state(ProcessState.WAITING, None, msg='Waiting to retrieve', data=RETRIEVE_COMMAND)

    def parse(self, retrieved_temporary_folder):
        """Return the `Running` state that will `parse` the `CalcJob`.

        :param retrieved_temporary_folder: temporary folder used in retrieving that can be used during parsing.
        """
        return self.create_state(ProcessState.RUNNING, self.process.parse, retrieved_temporary_folder)

    def interrupt(self, reason):
        """Interrupt the `Waiting` state by calling interrupt on the transport task `InterruptableFuture`."""
        if self._task is not None:
            self._task.interrupt(reason)

        if isinstance(reason, plumpy.KillInterruption):
            if self._killing is None:
                self._killing = plumpy.Future()
            return self._killing



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Correct the type string for the base data types

Revision ID: 0aebbeab274d
Revises: 7a6587e16f4c
Create Date: 2018-02-24 20:12:44.731358

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from alembic import op
from sqlalchemy.sql import text


# revision identifiers, used by Alembic.
revision = '0aebbeab274d'
down_revision = '7a6587e16f4c'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()

    # The base Data types Bool, Float, Int and Str have been moved in the source code, which means that their
    # module path changes, which determines the plugin type string which is stored in the databse.
    # The type string now will have a type string prefix that is unique to each sub type.
    statement = text("""
        UPDATE db_dbnode SET type = 'data.bool.Bool.' WHERE type = 'data.base.Bool.';
        UPDATE db_dbnode SET type = 'data.float.Float.' WHERE type = 'data.base.Float.';
        UPDATE db_dbnode SET type = 'data.int.Int.' WHERE type = 'data.base.Int.';
        UPDATE db_dbnode SET type = 'data.str.Str.' WHERE type = 'data.base.Str.';
        UPDATE db_dbnode SET type = 'data.list.List.' WHERE type = 'data.base.List.';
    """)
    conn.execute(statement)


def downgrade():
    conn = op.get_bind()

    statement = text("""
        UPDATE db_dbnode SET type = 'data.base.Bool.' WHERE type = 'data.bool.Bool.';
        UPDATE db_dbnode SET type = 'data.base.Float.' WHERE type = 'data.float.Float.';
        UPDATE db_dbnode SET type = 'data.base.Int.' WHERE type = 'data.int.Int.';
        UPDATE db_dbnode SET type = 'data.base.Str.' WHERE type = 'data.str.Str.';
        UPDATE db_dbnode SET type = 'data.base.List.' WHERE type = 'data.list.List.';
    """)
    conn.execute(statement)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Utilities for low-level querying of plugins.

Interface for operations that have to connect the
registry online and that operate on the cache folder.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import


def registry_cache_file_name():
    """
    return the name of the registry cache file
    """
    return 'plugins'


def registry_cache_file_path():
    """
    return the path to the cache file
    """
    from aiida.plugins.utils import registry_cache_folder_path
    from os import path as osp
    cache_name = registry_cache_file_name()
    cache_dir = registry_cache_folder_path()
    return osp.join(cache_dir, cache_name)


def registry_cache_exists():
    """
    check if the registry cache exists

    :return bool: True if exists, False if not
    """
    from os import path as osp
    return osp.exists(registry_cache_file_path())


def registry_cache_openable():
    """
    return true if the registry cache file can be opened
    """
    from aiida.plugins.utils import registry_cache_folder_exists
    if not registry_cache_folder_exists():
        return False
    elif not registry_cache_exists():
        return False
    else:
        try:
            with open(registry_cache_file_path()):
                return True
        except Exception as e:
            return False


def registry_file_url():
    """
    return the url for the plugins.json file
    """
    return 'https://raw.github.com/aiidateam/aiida-registry/master/plugins.json'


def load_online(errorhandler=None):
    """
    loads the registry file and returns the list of plugins
    """
    from aiida.plugins.utils import load_json_from_url
    import requests
    url = registry_file_url()
    return load_json_from_url(url, errorhandler=errorhandler)


def update(with_info=True, registry_err_handler=None, info_err_handler=None):
    """
    Load the registry from its online location and pickle it.

    Creates the cache file if necessary.
    By default updates the entry details cache for each entry as well.

    :param with_info: default: True, update info cache for each entry as well
    :param registry_err_handler: callable(exception) -> dict. Must either raise
        or return a registry dict
    :param info_err_handler: callable(exception, plugin, data) -> None. Can
        raise or just print an error / warning.

    If none of the error handlers are given, the function will stop execution if
    any broken links are encountered.
    """
    from aiida.plugins.utils import pickle_to_registry_cache_folder
    cache_fname = registry_cache_file_name()
    pluginlist = load_online(errorhandler=registry_err_handler)
    pickle_to_registry_cache_folder(pluginlist, cache_fname)
    if with_info:
        cleanup_info(pluginlist)
        update_info(pluginlist, errorhandler=info_err_handler)


def load_cached():
    """
    load the registry from the local cache
    if the local cache is not readable, create or update it
    """
    from six.moves.cPickle import load as pload
    from aiida.plugins.utils import unpickle_from_registry_cache_folder
    cache_fname = registry_cache_file_name()
    if not registry_cache_openable():
        update()
    return unpickle_from_registry_cache_folder(cache_fname)


def update_info(registry=None, errorhandler=None):
    """
    iterate through plugins, download setup info and return as dict
    """
    from aiida.plugins.utils import pickle_to_registry_cache_folder
    from aiida.plugins.entry import RegistryEntry as Entry

    if not registry:
        registry = load_cached()

    for plugin, data in registry.items():
        try:
            entry = Entry(**data)
            pickle_to_registry_cache_folder(entry, plugin)
        except Exception as e:
            if errorhandler:
                errorhandler(e, plugin, data)
            else:
                raise e


def cleanup_info(registry=None):
    """
    delete any plugin info files that do not correspond to a registry entry
    """
    from os import remove, listdir
    from os import path as osp
    from aiida.plugins.utils import registry_cache_folder_path

    if not registry:
        registry = load_cached()

    all_cache_files = listdir(registry_cache_folder_path())
    all_cache_files.remove(registry_cache_file_name())
    plugin_cache_files = all_cache_files
    cache_dir = registry_cache_folder_path()
    for cache_file in plugin_cache_files:
        if cache_file not in registry:
            remove(osp.join(cache_dir, cache_file))




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=global-statement
"""Definition of AiiDA's process persister and the necessary object loaders."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import logging
import traceback

import plumpy

from aiida.orm.utils import serialize

__all__ = ('AiiDAPersister', 'ObjectLoader', 'get_object_loader')

LOGGER = logging.getLogger(__name__)
OBJECT_LOADER = None

ObjectLoader = plumpy.DefaultObjectLoader


def get_object_loader():
    """
    Get the global AiiDA object loader

    :return: The global object loader
    :rtype: :class:`plumpy.ObjectLoader`
    """
    global OBJECT_LOADER
    if OBJECT_LOADER is None:
        OBJECT_LOADER = ObjectLoader()
    return OBJECT_LOADER


class AiiDAPersister(plumpy.Persister):
    """
    This node is responsible to taking saved process instance states and
    persisting them to the database.
    """

    def save_checkpoint(self, process, tag=None):
        """
        Persist a Process instance

        :param process: :class:`aiida.engine.Process`
        :param tag: optional checkpoint identifier to allow distinguishing multiple checkpoints for the same process
        :raises: :class:`plumpy.PersistenceError` Raised if there was a problem saving the checkpoint
        """
        LOGGER.debug('Persisting process<%d>', process.pid)

        if tag is not None:
            raise NotImplementedError('Checkpoint tags not supported yet')

        try:
            bundle = plumpy.Bundle(process, plumpy.LoadSaveContext(loader=get_object_loader()))
        except ValueError:
            # Couldn't create the bundle
            raise plumpy.PersistenceError(
                "Failed to create a bundle for '{}': {}".format(process, traceback.format_exc())
            )

        try:
            process.node.set_checkpoint(serialize.serialize(bundle))
        except Exception:
            raise plumpy.PersistenceError(
                "Failed to store a checkpoint for '{}': {}".format(process, traceback.format_exc())
            )

        return bundle

    def load_checkpoint(self, pid, tag=None):
        """
        Load a process from a persisted checkpoint by its process id

        :param pid: the process id of the :class:`plumpy.Process`
        :param tag: optional checkpoint identifier to allow retrieving a specific sub checkpoint
        :return: a bundle with the process state
        :rtype: :class:`plumpy.Bundle`
        :raises: :class:`plumpy.PersistenceError` Raised if there was a problem loading the checkpoint
        """
        from aiida.common.exceptions import MultipleObjectsError, NotExistent
        from aiida.orm import load_node

        if tag is not None:
            raise NotImplementedError('Checkpoint tags not supported yet')

        try:
            calculation = load_node(pid)
        except (MultipleObjectsError, NotExistent):
            raise plumpy.PersistenceError(
                'Failed to load the node for process<{}>: {}'.format(pid, traceback.format_exc())
            )

        checkpoint = calculation.checkpoint

        if checkpoint is None:
            raise plumpy.PersistenceError('Calculation<{}> does not have a saved checkpoint'.format(calculation.pk))

        try:
            bundle = serialize.deserialize(checkpoint)
        except Exception:
            raise plumpy.PersistenceError(
                'Failed to load the checkpoint for process<{}>: {}'.format(pid, traceback.format_exc())
            )

        return bundle

    def get_checkpoints(self):
        """
        Return a list of all the current persisted process checkpoints
        with each element containing the process id and optional checkpoint tag

        :return: list of PersistedCheckpoint tuples
        """

    def get_process_checkpoints(self, pid):
        """
        Return a list of all the current persisted process checkpoints for the
        specified process with each element containing the process id and
        optional checkpoint tag

        :param pid: the process pid
        :return: list of PersistedCheckpoint tuples
        """

    def delete_checkpoint(self, pid, tag=None):
        """
        Delete a persisted process checkpoint, where no error will be raised if the checkpoint does not exist

        :param pid: the process id of the :class:`plumpy.Process`
        :param tag: optional checkpoint identifier to allow retrieving a specific sub checkpoint
        """
        from aiida.orm import load_node

        calc = load_node(pid)
        calc.delete_checkpoint()

    def delete_process_checkpoints(self, pid):
        """
        Delete all persisted checkpoints related to the given process id

        :param pid: the process id of the :class:`aiida.engine.processes.process.Process`
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Plugin for PBSPro.
This has been tested on PBSPro v. 12.
"""
from __future__ import print_function
from __future__ import division

from __future__ import absolute_import
import logging
from .pbsbaseclasses import PbsBaseClass

_LOGGER = logging.getLogger(__name__)

# This maps PbsPro status letters to our own status list

## List of states from the man page of qstat
# B  Array job has at least one subjob running.
# E  Job is exiting after having run.
# F  Job is finished.
# H  Job is held.
# M  Job was moved to another server.
# Q  Job is queued.
# R  Job is running.
# S  Job is suspended.
# T  Job is being moved to new location.
# U  Cycle-harvesting job is suspended due to  keyboard  activity.
# W  Job is waiting for its submitter-assigned start time to be reached.
# X  Subjob has completed execution or has been deleted.


class PbsproScheduler(PbsBaseClass):
    """
    Subclass to support the PBSPro scheduler
    (http://www.pbsworks.com/).

    I redefine only what needs to change from the base class
    """

    ## I don't need to change this from the base class
    # _job_resource_class = PbsJobResource

    ## For the time being I use a common dictionary, should be sufficient
    ## for the time being, but I can redefine it if needed.
    # _map_status = _map_status_pbs_common

    def _get_resource_lines(
        self, num_machines, num_mpiprocs_per_machine, num_cores_per_machine, max_memory_kb, max_wallclock_seconds
    ):
        """
        Return the lines for machines, memory and wallclock relative
        to pbspro.
        """
        # Note: num_cores_per_machine is not used here but is provided by
        #       the parent class ('_get_submit_script_header') method

        return_lines = []

        select_string = 'select={}'.format(num_machines)
        if num_mpiprocs_per_machine:
            select_string += ':mpiprocs={}'.format(num_mpiprocs_per_machine)
        if num_cores_per_machine:
            select_string += ':ppn={}'.format(num_cores_per_machine)

        if max_wallclock_seconds is not None:
            try:
                tot_secs = int(max_wallclock_seconds)
                if tot_secs <= 0:
                    raise ValueError
            except ValueError:
                raise ValueError(
                    'max_wallclock_seconds must be '
                    "a positive integer (in seconds)! It is instead '{}'"
                    ''.format(max_wallclock_seconds)
                )
            hours = tot_secs // 3600
            tot_minutes = tot_secs % 3600
            minutes = tot_minutes // 60
            seconds = tot_minutes % 60
            return_lines.append('#PBS -l walltime={:02d}:{:02d}:{:02d}'.format(hours, minutes, seconds))

        if max_memory_kb:
            try:
                virtual_memory_kb = int(max_memory_kb)
                if virtual_memory_kb <= 0:
                    raise ValueError
            except ValueError:
                raise ValueError(
                    'max_memory_kb must be '
                    "a positive integer (in kB)! It is instead '{}'"
                    ''.format((max_memory_kb))
                )
            select_string += ':mem={}kb'.format(virtual_memory_kb)

        return_lines.append('#PBS -l {}'.format(select_string))
        return return_lines
